"""
Smart query generation strategy that works for any type of search target.
"""

import concurrent.futures
from typing import Dict, List

from loguru import logger

from ..constraints.base_constraint import Constraint
from ..constraints.constraint_analyzer import ConstraintType
from .early_stop_constrained_strategy import EarlyStopConstrainedStrategy


class SmartQueryStrategy(EarlyStopConstrainedStrategy):
    """
    Enhanced strategy with intelligent query generation that:
    1. Analyzes constraints to identify key search terms
    2. Uses LLM to suggest search queries based on constraint meaning
    3. Generates multiple query variations for better coverage
    """

    def __init__(
        self,
        *args,
        use_llm_query_generation: bool = True,
        queries_per_combination: int = 3,
        use_entity_seeding: bool = True,
        use_direct_property_search: bool = True,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.use_llm_query_generation = use_llm_query_generation
        self.queries_per_combination = queries_per_combination
        self.use_entity_seeding = use_entity_seeding
        self.use_direct_property_search = use_direct_property_search
        # Track queries to avoid duplicates
        self.searched_queries = set()
        self.query_variations = set()
        # Store entity seeds for targeted search
        self.entity_seeds = []

    def _build_query(self, constraints: List[Constraint]) -> str:
        """Build intelligent queries using constraint analysis."""
        if self.use_llm_query_generation:
            # Use LLM to generate smart queries
            return self._generate_smart_query(constraints)
        else:
            # Fallback to improved standard approach
            return self._build_standard_query(constraints)

    def _generate_smart_query(self, constraints: List[Constraint]) -> str:
        """Use LLM to generate optimal search queries."""
        constraint_text = "\n".join(
            [
                f"- {c.type.value}: {c.value} (weight: {c.weight})"
                for c in constraints
            ]
        )

        # Build a list of already searched queries to avoid duplication
        searched_list = list(self.searched_queries)[:10]  # Show last 10 to LLM
        already_searched = (
            "\n".join([f"- {q}" for q in searched_list])
            if searched_list
            else "None"
        )

        prompt = f"""
Analyze these search constraints and generate an optimal web search query:

Constraints:
{constraint_text}

Target type: {getattr(self, "entity_type", "unknown")}

Already searched queries (avoid these):
{already_searched}

Generate a single search query that would most effectively find results matching these constraints.
The query should:
1. Include the most identifying/unique terms
2. Use appropriate search operators (quotes, AND, OR)
3. Be specific enough to find relevant results but not too narrow
4. Focus on the highest weighted constraints
5. Be different from already searched queries

Return only the search query, nothing else.
"""

        try:
            query = self.model.invoke(prompt).content.strip()

            # Check if this query is too similar to existing ones
            normalized_query = query.strip().lower()
            if normalized_query in self.searched_queries:
                logger.info(
                    f"LLM generated duplicate query, using fallback: {query}"
                )
                return self._build_standard_query(constraints)

            logger.info(f"LLM generated query: {query}")
            return query
        except Exception:
            logger.exception("Failed to generate smart query")
            return self._build_standard_query(constraints)

    def _build_standard_query(self, constraints: List[Constraint]) -> str:
        """Improved standard query building."""
        # Group constraints by importance
        critical_terms = []
        supplementary_terms = []

        for c in constraints:
            term = c.value

            # Quote multi-word terms
            if " " in term and not term.startswith('"'):
                term = f'"{term}"'

            if c.weight > 0.7:
                critical_terms.append(term)
            else:
                supplementary_terms.append(term)

        # Build query with critical terms required, supplementary optional
        query_parts = []

        # Add entity type if known
        entity_type = getattr(self, "entity_type", None)
        if entity_type and entity_type != "unknown entity":
            query_parts.append(entity_type)

        # Add critical terms
        if critical_terms:
            query_parts.extend(critical_terms)

        # Add some supplementary terms
        if supplementary_terms:
            query_parts.extend(
                supplementary_terms[:2]
            )  # Limit to avoid overly specific queries

        return " ".join(query_parts)

    def _execute_combination_search(self, combo) -> List:
        """Override to generate multiple query variations per combination."""
        all_candidates = []

        if self.use_llm_query_generation:
            # Generate multiple query variations
            queries = self._generate_query_variations(combo.constraints)

            # Execute searches in parallel
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=self.queries_per_combination
            ) as executor:
                futures = []
                for query in queries:
                    # Check if we've already searched this query
                    normalized_query = query.strip().lower()
                    if normalized_query in self.searched_queries:
                        logger.info(f"Skipping duplicate query: '{query}'")
                        continue

                    self.searched_queries.add(normalized_query)
                    future = executor.submit(self._execute_search, query)
                    futures.append((query, future))

                for query, future in futures:
                    try:
                        results = future.result()
                        candidates = self._extract_candidates_from_results(
                            results
                        )
                        all_candidates.extend(candidates)

                        logger.info(
                            f"Query '{query}' found {len(candidates)} candidates"
                        )
                    except Exception as e:
                        logger.exception(
                            f"Search failed for query '{query}': {e}"
                        )
        else:
            # Use single query from parent implementation
            candidates = super()._execute_combination_search(combo)
            all_candidates.extend(candidates)

        return all_candidates

    def _generate_query_variations(
        self, constraints: List[Constraint]
    ) -> List[str]:
        """Generate multiple query variations for better coverage."""
        # Handle single constraint case
        if isinstance(constraints, Constraint):
            constraints = [constraints]

        constraint_text = "\n".join(
            [f"- {c.type.value}: {c.value}" for c in constraints]
        )

        # Build a list of already searched queries to avoid duplication
        searched_list = list(self.searched_queries)[:20]  # Show last 20 to LLM
        already_searched = (
            "\n".join([f"- {q}" for q in searched_list])
            if searched_list
            else "None"
        )

        prompt = f"""
Generate {self.queries_per_combination} different search queries for these constraints:

{constraint_text}

Already searched queries (avoid these):
{already_searched}

Each query should:
- Approach the search from a different angle
- Use different search terms or operators
- Target different aspects of the constraints
- Be distinctly different from already searched queries

Provide each query on a separate line.
"""

        try:
            response = self.model.invoke(prompt).content
            queries = [q.strip() for q in response.split("\n") if q.strip()]

            # Filter out duplicates
            unique_queries = []
            for query in queries:
                normalized = query.strip().lower()
                if (
                    normalized not in self.searched_queries
                    and normalized not in self.query_variations
                ):
                    unique_queries.append(query)
                    self.query_variations.add(normalized)
                else:
                    logger.info(
                        f"Filtering out duplicate query variation: {query}"
                    )

            # If all queries were duplicates, generate a fallback
            if not unique_queries:
                fallback = self._build_standard_query(constraints)
                if fallback.strip().lower() not in self.searched_queries:
                    unique_queries = [fallback]

            return unique_queries[: self.queries_per_combination]
        except Exception:
            logger.exception("Failed to generate query variations")
            # Fallback to single query
            return [self._build_standard_query(constraints)]

    def _extract_candidates_from_results(self, results: Dict) -> List:
        """Improved candidate extraction that's more generic."""
        candidates = []
        content = results.get("current_knowledge", "")

        if not content:
            return candidates

        # Use LLM to extract relevant entities/topics from the content
        prompt = f"""
From the following search results, extract all relevant entities, topics, or answers that match our search target type: {getattr(self, "entity_type", "unknown")}

Content:
{content}

List each potential match on a separate line.
Include only names/titles/identifiers, not descriptions.
"""

        try:
            response = self.model.invoke(prompt).content
            entity_names = [
                name.strip() for name in response.split("\n") if name.strip()
            ]

            # Create candidates from extracted names
            from ..candidates.base_candidate import Candidate

            for name in entity_names:
                if name and len(name) < 100:  # Basic validation
                    candidate = Candidate(name=name)
                    candidates.append(candidate)

            logger.info(f"Extracted {len(candidates)} candidates from results")

        except Exception:
            logger.exception("Error extracting candidates")

        return candidates

    def _should_use_entity_seeding(self) -> bool:
        """Determine if entity seeding would be beneficial."""
        entity_type = getattr(self, "entity_type", "").lower()
        return (
            "character" in entity_type
            or "person" in entity_type
            or "hero" in entity_type
        )

    def _perform_entity_seeding(self):
        """Use LLM to suggest specific entity names based on constraints."""
        logger.info("Performing entity seeding based on constraints")

        # Extract key properties from constraints
        key_properties = []
        for constraint in self.constraint_ranking:
            if constraint.weight > 0.7:  # High-weight constraints
                key_properties.append(constraint.value)

        if not key_properties:
            return

        properties_text = "\n".join([f"- {prop}" for prop in key_properties])

        prompt = f"""
Based on these properties, suggest 5-10 specific {self.entity_type} names that might match:

Properties:
{properties_text}

For example, if looking for a scientist from the 19th century, you might suggest:
- Charles Darwin
- Marie Curie
- Louis Pasteur
- Thomas Edison

Provide one name per line. Be specific with actual character/entity names.
"""

        try:
            response = self.model.invoke(prompt).content
            self.entity_seeds = [
                name.strip() for name in response.split("\n") if name.strip()
            ]
            logger.info(f"Generated entity seeds: {self.entity_seeds}")

            # Immediately search for these seeds
            self._search_entity_seeds()

        except Exception:
            logger.exception("Error generating entity seeds")

    def _search_entity_seeds(self):
        """Search for the entity seeds directly."""
        if not self.entity_seeds:
            return

        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            futures = []
            for seed in self.entity_seeds[:5]:  # Limit to top 5
                query = f'"{seed}"'
                if query.lower() not in self.searched_queries:
                    self.searched_queries.add(query.lower())
                    future = executor.submit(self._execute_search, query)
                    futures.append((seed, future))

            for seed, future in futures:
                try:
                    results = future.result()
                    candidates = self._extract_candidates_from_results(results)

                    # Look for exact matches
                    for candidate in candidates:
                        if seed.lower() in candidate.name.lower():
                            logger.info(
                                f"Found seeded entity: {candidate.name}"
                            )
                            # Evaluate immediately
                            if hasattr(self, "_evaluate_candidate_immediately"):
                                self._evaluate_candidate_immediately(candidate)
                            else:
                                # Add to candidates list
                                if not hasattr(self, "candidates"):
                                    self.candidates = []
                                self.candidates.append(candidate)

                except Exception:
                    logger.exception(f"Error searching for seed {seed}")

    def _try_direct_property_search(self):
        """Try direct searches for high-weight property constraints."""
        property_queries = []

        for constraint in self.constraint_ranking:
            if (
                constraint.weight > 0.7
                and constraint.type == ConstraintType.PROPERTY
            ):
                # Create specific property-based queries
                if (
                    "elastic" in constraint.value.lower()
                    or "stretch" in constraint.value.lower()
                ):
                    property_queries.extend(
                        [
                            f'"{constraint.value}" superhero character',
                            f'characters with "{constraint.value}"',
                            f"list of {self.entity_type} {constraint.value}",
                        ]
                    )
                elif (
                    "voice" in constraint.value.lower()
                    or "actor" in constraint.value.lower()
                ):
                    property_queries.append(
                        f"{constraint.value} {self.entity_type}"
                    )

        # Execute property searches
        if property_queries:
            logger.info(
                f"Executing direct property searches: {property_queries}"
            )
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=3
            ) as executor:
                futures = []
                for query in property_queries[
                    :3
                ]:  # Limit to avoid too many searches
                    if query.lower() not in self.searched_queries:
                        self.searched_queries.add(query.lower())
                        future = executor.submit(self._execute_search, query)
                        futures.append(future)

                for future in futures:
                    try:
                        results = future.result()
                        candidates = self._extract_candidates_from_results(
                            results
                        )

                        for candidate in candidates:
                            if hasattr(self, "_evaluate_candidate_immediately"):
                                self._evaluate_candidate_immediately(candidate)

                    except Exception:
                        logger.exception("Property search error")

    def _perform_entity_name_search(self):
        """Last resort: search for entity names directly with constraints."""
        logger.info("Performing entity name search fallback")

        for entity_name in self.entity_seeds[:3]:  # Top 3 seeds
            # Combine entity name with key constraints
            constraint_terms = []
            for constraint in self.constraint_ranking[:2]:  # Top 2 constraints
                if constraint.weight > 0.5:
                    constraint_terms.append(constraint.value)

            if constraint_terms:
                query = f'"{entity_name}" {" ".join(constraint_terms)}'
                if query.lower() not in self.searched_queries:
                    logger.info(f"Trying targeted entity search: {query}")
                    self.searched_queries.add(query.lower())

                    try:
                        results = self._execute_search(query)
                        candidates = self._extract_candidates_from_results(
                            results
                        )

                        for candidate in candidates:
                            if entity_name.lower() in candidate.name.lower():
                                logger.info(
                                    f"Found target entity in fallback: {candidate.name}"
                                )
                                if hasattr(
                                    self, "_evaluate_candidate_immediately"
                                ):
                                    self._evaluate_candidate_immediately(
                                        candidate
                                    )

                                    # Check for early stop
                                    if (
                                        hasattr(self, "best_score")
                                        and self.best_score >= 0.99
                                    ):
                                        return

                    except Exception:
                        logger.exception("Entity name search error")

    def _progressive_constraint_search(self):
        """Override to add entity seeding and property search."""
        # Detect entity type first
        self.entity_type = self._detect_entity_type()
        logger.info(f"Detected entity type: {self.entity_type}")

        # Perform entity seeding if enabled and entity type suggests specific entities
        if self.use_entity_seeding and self._should_use_entity_seeding():
            self._perform_entity_seeding()

        # Try direct property search for high-weight properties
        if self.use_direct_property_search:
            self._try_direct_property_search()

        # Continue with normal progressive search
        super()._progressive_constraint_search()

        # If still no good results, try name-based fallback
        if (
            hasattr(self, "best_score")
            and self.best_score < 0.9
            and self.entity_seeds
        ):
            self._perform_entity_name_search()



"""Tests for Direct PDF downloader."""

from unittest.mock import Mock, patch


from local_deep_research.research_library.downloaders.direct_pdf import (
    DirectPDFDownloader,
)


class TestDirectPDFDownloaderInit:
    """Tests for DirectPDFDownloader initialization."""

    def test_initializes_with_default_timeout(self):
        """Should initialize with default timeout."""
        downloader = DirectPDFDownloader()
        assert downloader.timeout == 30

    def test_initializes_with_custom_timeout(self):
        """Should accept custom timeout."""
        downloader = DirectPDFDownloader(timeout=60)
        assert downloader.timeout == 60

    def test_creates_session(self):
        """Should create a requests session."""
        downloader = DirectPDFDownloader()
        assert downloader.session is not None


class TestDirectPDFDownloaderCanHandle:
    """Tests for can_handle method."""

    def test_handles_pdf_url(self):
        """Should handle URLs ending with .pdf."""
        downloader = DirectPDFDownloader()
        assert downloader.can_handle("https://example.com/paper.pdf")
        assert downloader.can_handle("https://example.com/path/to/document.PDF")

    def test_handles_pdf_url_with_query_params(self):
        """Should handle PDF URLs with query parameters."""
        downloader = DirectPDFDownloader()
        assert downloader.can_handle(
            "https://example.com/paper.pdf?token=abc123"
        )

    def test_rejects_non_pdf_urls(self):
        """Should reject non-PDF URLs."""
        downloader = DirectPDFDownloader()
        assert not downloader.can_handle("https://example.com/paper.html")
        assert not downloader.can_handle("https://example.com/document")
        assert not downloader.can_handle("https://arxiv.org/abs/1234.5678")

    def test_rejects_empty_url(self):
        """Should reject empty URLs."""
        downloader = DirectPDFDownloader()
        assert not downloader.can_handle("")


class TestDirectPDFDownloaderDownload:
    """Tests for download method."""

    def test_downloads_pdf_successfully(self, mock_pdf_content):
        """Should download PDF successfully."""
        downloader = DirectPDFDownloader()

        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.content = mock_pdf_content
        mock_response.headers = {"Content-Type": "application/pdf"}
        mock_response.raise_for_status = Mock()

        with patch.object(
            downloader.session, "get", return_value=mock_response
        ):
            result = downloader.download("https://example.com/paper.pdf")

        assert result is not None
        assert result == mock_pdf_content

    def test_handles_download_failure(self):
        """Should handle download failure gracefully."""
        downloader = DirectPDFDownloader()

        with patch.object(
            downloader.session, "get", side_effect=Exception("Network error")
        ):
            result = downloader.download("https://example.com/paper.pdf")

        assert result is None

    def test_handles_404_response(self):
        """Should handle 404 response."""
        downloader = DirectPDFDownloader()

        mock_response = Mock()
        mock_response.status_code = 404
        mock_response.raise_for_status = Mock(
            side_effect=Exception("404 Not Found")
        )

        with patch.object(
            downloader.session, "get", return_value=mock_response
        ):
            result = downloader.download("https://example.com/paper.pdf")

        assert result is None


class TestDirectPDFDownloaderDownloadWithResult:
    """Tests for download_with_result method."""

    def test_returns_download_result_on_success(self, mock_pdf_content):
        """Should return DownloadResult with content on success."""
        downloader = DirectPDFDownloader()

        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.content = mock_pdf_content
        mock_response.headers = {"Content-Type": "application/pdf"}
        mock_response.raise_for_status = Mock()

        with patch.object(
            downloader.session, "get", return_value=mock_response
        ):
            result = downloader.download_with_result(
                "https://example.com/paper.pdf"
            )

        assert result.is_success is True
        assert result.content == mock_pdf_content

    def test_returns_skip_reason_on_failure(self):
        """Should return skip_reason when download fails."""
        downloader = DirectPDFDownloader()

        with patch.object(
            downloader.session, "get", side_effect=Exception("Network error")
        ):
            result = downloader.download_with_result(
                "https://example.com/paper.pdf"
            )

        assert result.is_success is False
        assert result.skip_reason is not None



"""Tests for DocumentDeletionService."""

from unittest.mock import MagicMock, Mock, patch


from local_deep_research.research_library.deletion.services.document_deletion import (
    DocumentDeletionService,
)


class TestDocumentDeletionServiceInit:
    """Tests for DocumentDeletionService initialization."""

    def test_initializes_with_username(self):
        """Should initialize with username."""
        service = DocumentDeletionService(username="testuser")
        assert service.username == "testuser"


class TestDocumentDeletionServiceDeleteDocument:
    """Tests for delete_document method."""

    def test_returns_error_when_document_not_found(self):
        """Should return error when document doesn't exist."""
        service = DocumentDeletionService(username="testuser")

        with patch(
            "local_deep_research.research_library.deletion.services.document_deletion.get_user_db_session"
        ) as mock_get_session:
            mock_session = MagicMock()
            mock_cm = MagicMock()
            mock_cm.__enter__ = Mock(return_value=mock_session)
            mock_cm.__exit__ = Mock(return_value=None)
            mock_get_session.return_value = mock_cm

            mock_session.query.return_value.get.return_value = None

            result = service.delete_document("nonexistent-id")

        assert result["deleted"] is False
        assert "not found" in result["error"].lower()

    def test_deletes_document_successfully(self):
        """Should delete document and return stats."""
        service = DocumentDeletionService(username="testuser")

        with patch(
            "local_deep_research.research_library.deletion.services.document_deletion.get_user_db_session"
        ) as mock_get_session:
            mock_session = MagicMock()
            mock_cm = MagicMock()
            mock_cm.__enter__ = Mock(return_value=mock_session)
            mock_cm.__exit__ = Mock(return_value=None)
            mock_get_session.return_value = mock_cm

            # Mock document
            mock_doc = MagicMock()
            mock_doc.id = "doc-123"
            mock_doc.title = "Test Document"
            mock_doc.filename = "test.pdf"
            mock_doc.storage_mode = "database"
            mock_doc.file_path = None
            mock_session.query.return_value.get.return_value = mock_doc

            with patch(
                "local_deep_research.research_library.deletion.services.document_deletion.CascadeHelper"
            ) as mock_helper:
                mock_helper.get_document_collections.return_value = ["col-1"]
                mock_helper.delete_document_chunks.return_value = 5
                mock_helper.get_document_blob_size.return_value = 1024
                mock_helper.delete_document_completely.return_value = True

                result = service.delete_document("doc-123")

        assert result["deleted"] is True
        assert result["document_id"] == "doc-123"
        assert result["chunks_deleted"] == 5
        assert result["blob_size"] == 1024

    def test_handles_exception_gracefully(self):
        """Should handle exceptions and rollback."""
        service = DocumentDeletionService(username="testuser")

        with patch(
            "local_deep_research.research_library.deletion.services.document_deletion.get_user_db_session"
        ) as mock_get_session:
            mock_session = MagicMock()
            mock_cm = MagicMock()
            mock_cm.__enter__ = Mock(return_value=mock_session)
            mock_cm.__exit__ = Mock(return_value=None)
            mock_get_session.return_value = mock_cm

            mock_session.query.side_effect = Exception("DB Error")

            result = service.delete_document("doc-123")

        assert result["deleted"] is False
        assert "error" in result


class TestDocumentDeletionServiceDeleteBlobOnly:
    """Tests for delete_blob_only method."""

    def test_returns_error_when_document_not_found(self):
        """Should return error when document doesn't exist."""
        service = DocumentDeletionService(username="testuser")

        with patch(
            "local_deep_research.research_library.deletion.services.document_deletion.get_user_db_session"
        ) as mock_get_session:
            mock_session = MagicMock()
            mock_cm = MagicMock()
            mock_cm.__enter__ = Mock(return_value=mock_session)
            mock_cm.__exit__ = Mock(return_value=None)
            mock_get_session.return_value = mock_cm

            mock_session.query.return_value.get.return_value = None

            result = service.delete_blob_only("nonexistent-id")

        assert result["deleted"] is False
        assert "not found" in result["error"].lower()

    def test_deletes_database_blob(self):
        """Should delete blob from database storage."""
        service = DocumentDeletionService(username="testuser")

        with patch(
            "local_deep_research.research_library.deletion.services.document_deletion.get_user_db_session"
        ) as mock_get_session:
            mock_session = MagicMock()
            mock_cm = MagicMock()
            mock_cm.__enter__ = Mock(return_value=mock_session)
            mock_cm.__exit__ = Mock(return_value=None)
            mock_get_session.return_value = mock_cm

            mock_doc = MagicMock()
            mock_doc.id = "doc-123"
            mock_doc.storage_mode = "database"
            mock_session.query.return_value.get.return_value = mock_doc

            with patch(
                "local_deep_research.research_library.deletion.services.document_deletion.CascadeHelper"
            ) as mock_helper:
                mock_helper.delete_document_blob.return_value = 2048

                result = service.delete_blob_only("doc-123")

        assert result["deleted"] is True
        assert result["bytes_freed"] == 2048
        assert result["storage_mode_updated"] is True

    def test_returns_error_for_none_storage_mode(self):
        """Should return error when document has no stored PDF."""
        service = DocumentDeletionService(username="testuser")

        with patch(
            "local_deep_research.research_library.deletion.services.document_deletion.get_user_db_session"
        ) as mock_get_session:
            mock_session = MagicMock()
            mock_cm = MagicMock()
            mock_cm.__enter__ = Mock(return_value=mock_session)
            mock_cm.__exit__ = Mock(return_value=None)
            mock_get_session.return_value = mock_cm

            mock_doc = MagicMock()
            mock_doc.id = "doc-123"
            mock_doc.storage_mode = "none"
            mock_session.query.return_value.get.return_value = mock_doc

            result = service.delete_blob_only("doc-123")

        assert result["deleted"] is False
        assert "no stored pdf" in result["error"].lower()


class TestDocumentDeletionServiceRemoveFromCollection:
    """Tests for remove_from_collection method."""

    def test_returns_error_when_document_not_found(self):
        """Should return error when document doesn't exist."""
        service = DocumentDeletionService(username="testuser")

        with patch(
            "local_deep_research.research_library.deletion.services.document_deletion.get_user_db_session"
        ) as mock_get_session:
            mock_session = MagicMock()
            mock_cm = MagicMock()
            mock_cm.__enter__ = Mock(return_value=mock_session)
            mock_cm.__exit__ = Mock(return_value=None)
            mock_get_session.return_value = mock_cm

            mock_session.query.return_value.get.return_value = None

            result = service.remove_from_collection("doc-123", "col-456")

        assert result["unlinked"] is False
        assert "not found" in result["error"].lower()

    def test_returns_error_when_not_in_collection(self):
        """Should return error when document not in collection."""
        service = DocumentDeletionService(username="testuser")

        with patch(
            "local_deep_research.research_library.deletion.services.document_deletion.get_user_db_session"
        ) as mock_get_session:
            mock_session = MagicMock()
            mock_cm = MagicMock()
            mock_cm.__enter__ = Mock(return_value=mock_session)
            mock_cm.__exit__ = Mock(return_value=None)
            mock_get_session.return_value = mock_cm

            mock_doc = MagicMock()
            mock_session.query.return_value.get.return_value = mock_doc
            mock_session.query.return_value.filter_by.return_value.first.return_value = None

            result = service.remove_from_collection("doc-123", "col-456")

        assert result["unlinked"] is False
        assert "not in this collection" in result["error"].lower()

    def test_unlinks_document_from_collection(self):
        """Should unlink document from collection."""
        service = DocumentDeletionService(username="testuser")

        with patch(
            "local_deep_research.research_library.deletion.services.document_deletion.get_user_db_session"
        ) as mock_get_session:
            mock_session = MagicMock()
            mock_cm = MagicMock()
            mock_cm.__enter__ = Mock(return_value=mock_session)
            mock_cm.__exit__ = Mock(return_value=None)
            mock_get_session.return_value = mock_cm

            mock_doc = MagicMock()
            mock_doc.id = "doc-123"
            mock_doc_collection = MagicMock()

            # Set up query chain
            mock_session.query.return_value.get.return_value = mock_doc
            mock_session.query.return_value.filter_by.return_value.first.return_value = mock_doc_collection

            with patch(
                "local_deep_research.research_library.deletion.services.document_deletion.CascadeHelper"
            ) as mock_helper:
                mock_helper.delete_document_chunks.return_value = 3
                mock_helper.count_document_in_collections.return_value = (
                    1  # Still in another collection
                )

                result = service.remove_from_collection("doc-123", "col-456")

        assert result["unlinked"] is True
        assert result["chunks_deleted"] == 3
        assert result["document_deleted"] is False


class TestDocumentDeletionServiceGetDeletionPreview:
    """Tests for get_deletion_preview method."""

    def test_returns_not_found_for_missing_document(self):
        """Should return found=False for missing document."""
        service = DocumentDeletionService(username="testuser")

        with patch(
            "local_deep_research.research_library.deletion.services.document_deletion.get_user_db_session"
        ) as mock_get_session:
            mock_session = MagicMock()
            mock_cm = MagicMock()
            mock_cm.__enter__ = Mock(return_value=mock_session)
            mock_cm.__exit__ = Mock(return_value=None)
            mock_get_session.return_value = mock_cm

            mock_session.query.return_value.get.return_value = None

            result = service.get_deletion_preview("nonexistent-id")

        assert result["found"] is False

    def test_returns_document_details(self):
        """Should return document details for preview."""
        service = DocumentDeletionService(username="testuser")

        with patch(
            "local_deep_research.research_library.deletion.services.document_deletion.get_user_db_session"
        ) as mock_get_session:
            mock_session = MagicMock()
            mock_cm = MagicMock()
            mock_cm.__enter__ = Mock(return_value=mock_session)
            mock_cm.__exit__ = Mock(return_value=None)
            mock_get_session.return_value = mock_cm

            mock_doc = MagicMock()
            mock_doc.id = "doc-123"
            mock_doc.title = "Test Document"
            mock_doc.filename = "test.pdf"
            mock_doc.file_type = "pdf"
            mock_doc.storage_mode = "database"
            mock_doc.text_content = "Some text content"
            mock_session.query.return_value.get.return_value = mock_doc
            mock_session.query.return_value.filter.return_value.count.return_value = 10

            with patch(
                "local_deep_research.research_library.deletion.services.document_deletion.CascadeHelper"
            ) as mock_helper:
                mock_helper.get_document_collections.return_value = [
                    "col-1",
                    "col-2",
                ]
                mock_helper.get_document_blob_size.return_value = 5120

                result = service.get_deletion_preview("doc-123")

        assert result["found"] is True
        assert result["title"] == "Test Document"
        assert result["collections_count"] == 2
        assert result["blob_size"] == 5120



"""
Comprehensive tests for search_utilities module.
Tests tag removal, link extraction, markdown formatting, and findings formatting.
"""

from unittest.mock import patch


class TestLanguageCodeMap:
    """Tests for LANGUAGE_CODE_MAP constant."""

    def test_language_code_map_exists(self):
        """Test that LANGUAGE_CODE_MAP is defined."""
        from local_deep_research.utilities.search_utilities import (
            LANGUAGE_CODE_MAP,
        )

        assert LANGUAGE_CODE_MAP is not None
        assert isinstance(LANGUAGE_CODE_MAP, dict)

    def test_contains_common_languages(self):
        """Test that common languages are mapped."""
        from local_deep_research.utilities.search_utilities import (
            LANGUAGE_CODE_MAP,
        )

        assert LANGUAGE_CODE_MAP["english"] == "en"
        assert LANGUAGE_CODE_MAP["french"] == "fr"
        assert LANGUAGE_CODE_MAP["german"] == "de"
        assert LANGUAGE_CODE_MAP["spanish"] == "es"

    def test_contains_asian_languages(self):
        """Test that Asian languages are mapped."""
        from local_deep_research.utilities.search_utilities import (
            LANGUAGE_CODE_MAP,
        )

        assert LANGUAGE_CODE_MAP["japanese"] == "ja"
        assert LANGUAGE_CODE_MAP["chinese"] == "zh"


class TestRemoveThinkTags:
    """Tests for remove_think_tags function."""

    def test_removes_paired_think_tags(self):
        """Test removal of paired <think>...</think> tags."""
        from local_deep_research.utilities.search_utilities import (
            remove_think_tags,
        )

        text = "Hello <think>internal thoughts</think> World"
        result = remove_think_tags(text)

        assert result == "Hello  World"
        assert "<think>" not in result
        assert "</think>" not in result

    def test_removes_multiline_think_tags(self):
        """Test removal of multiline think tags."""
        from local_deep_research.utilities.search_utilities import (
            remove_think_tags,
        )

        text = "Start <think>\nLine 1\nLine 2\n</think> End"
        result = remove_think_tags(text)

        assert result == "Start  End"

    def test_removes_multiple_think_tags(self):
        """Test removal of multiple think tag pairs."""
        from local_deep_research.utilities.search_utilities import (
            remove_think_tags,
        )

        text = "A <think>x</think> B <think>y</think> C"
        result = remove_think_tags(text)

        assert result == "A  B  C"

    def test_removes_orphaned_opening_tag(self):
        """Test removal of orphaned opening think tag."""
        from local_deep_research.utilities.search_utilities import (
            remove_think_tags,
        )

        text = "Hello <think> World"
        result = remove_think_tags(text)

        assert "<think>" not in result

    def test_removes_orphaned_closing_tag(self):
        """Test removal of orphaned closing think tag."""
        from local_deep_research.utilities.search_utilities import (
            remove_think_tags,
        )

        text = "Hello </think> World"
        result = remove_think_tags(text)

        assert "</think>" not in result

    def test_strips_whitespace(self):
        """Test that result is stripped of leading/trailing whitespace."""
        from local_deep_research.utilities.search_utilities import (
            remove_think_tags,
        )

        text = "  <think>thoughts</think>  Content  "
        result = remove_think_tags(text)

        assert result == "Content"

    def test_handles_empty_think_tags(self):
        """Test handling of empty think tags."""
        from local_deep_research.utilities.search_utilities import (
            remove_think_tags,
        )

        text = "Hello <think></think> World"
        result = remove_think_tags(text)

        assert result == "Hello  World"

    def test_handles_text_without_think_tags(self):
        """Test that text without think tags is unchanged."""
        from local_deep_research.utilities.search_utilities import (
            remove_think_tags,
        )

        text = "Hello World"
        result = remove_think_tags(text)

        assert result == "Hello World"

    def test_handles_empty_string(self):
        """Test handling of empty string."""
        from local_deep_research.utilities.search_utilities import (
            remove_think_tags,
        )

        result = remove_think_tags("")

        assert result == ""


class TestExtractLinksFromSearchResults:
    """Tests for extract_links_from_search_results function."""

    def test_extracts_links_from_valid_results(self, sample_search_results):
        """Test extraction from valid search results."""
        from local_deep_research.utilities.search_utilities import (
            extract_links_from_search_results,
        )

        links = extract_links_from_search_results(sample_search_results)

        assert len(links) == 3
        assert links[0]["title"] == "First Article"
        assert links[0]["url"] == "https://example.com/article1"

    def test_returns_empty_list_for_none(self):
        """Test that None input returns empty list."""
        from local_deep_research.utilities.search_utilities import (
            extract_links_from_search_results,
        )

        links = extract_links_from_search_results(None)

        assert links == []

    def test_returns_empty_list_for_empty_list(self):
        """Test that empty list returns empty list."""
        from local_deep_research.utilities.search_utilities import (
            extract_links_from_search_results,
        )

        links = extract_links_from_search_results([])

        assert links == []

    def test_extracts_index(self, sample_search_results):
        """Test that index is extracted."""
        from local_deep_research.utilities.search_utilities import (
            extract_links_from_search_results,
        )

        links = extract_links_from_search_results(sample_search_results)

        assert links[0]["index"] == "1"
        assert links[1]["index"] == "2"

    def test_handles_none_title(self):
        """Test handling of None title."""
        from local_deep_research.utilities.search_utilities import (
            extract_links_from_search_results,
        )

        results = [{"title": None, "link": "https://example.com"}]
        links = extract_links_from_search_results(results)

        # Should skip results with None title
        assert links == []

    def test_handles_none_link(self):
        """Test handling of None link."""
        from local_deep_research.utilities.search_utilities import (
            extract_links_from_search_results,
        )

        results = [{"title": "Test", "link": None}]
        links = extract_links_from_search_results(results)

        # Should skip results with None link
        assert links == []

    def test_handles_missing_title(self):
        """Test handling of missing title key."""
        from local_deep_research.utilities.search_utilities import (
            extract_links_from_search_results,
        )

        results = [{"link": "https://example.com"}]
        links = extract_links_from_search_results(results)

        # Should skip results without title
        assert links == []

    def test_handles_missing_link(self):
        """Test handling of missing link key."""
        from local_deep_research.utilities.search_utilities import (
            extract_links_from_search_results,
        )

        results = [{"title": "Test"}]
        links = extract_links_from_search_results(results)

        # Should skip results without link
        assert links == []

    def test_strips_whitespace_from_values(self):
        """Test that whitespace is stripped from values."""
        from local_deep_research.utilities.search_utilities import (
            extract_links_from_search_results,
        )

        results = [
            {"title": "  Test Title  ", "link": "  https://example.com  "}
        ]
        links = extract_links_from_search_results(results)

        assert links[0]["title"] == "Test Title"
        assert links[0]["url"] == "https://example.com"

    def test_handles_missing_index(self):
        """Test handling of missing index key."""
        from local_deep_research.utilities.search_utilities import (
            extract_links_from_search_results,
        )

        results = [{"title": "Test", "link": "https://example.com"}]
        links = extract_links_from_search_results(results)

        assert links[0]["index"] == ""

    def test_skips_invalid_results(self):
        """Test that invalid results are skipped."""
        from local_deep_research.utilities.search_utilities import (
            extract_links_from_search_results,
        )

        results = [
            {"title": "Valid", "link": "https://example.com"},
            {"title": "", "link": ""},  # Empty - should skip
            {"title": "Also Valid", "link": "https://test.com"},
        ]
        links = extract_links_from_search_results(results)

        assert len(links) == 2


class TestFormatLinksToMarkdown:
    """Tests for format_links_to_markdown function."""

    def test_formats_links_with_indices(self):
        """Test formatting links with indices."""
        from local_deep_research.utilities.search_utilities import (
            format_links_to_markdown,
        )

        links = [
            {
                "title": "Test Article",
                "url": "https://example.com",
                "index": "1",
            }
        ]
        result = format_links_to_markdown(links)

        assert "[1]" in result
        assert "Test Article" in result
        assert "https://example.com" in result

    def test_deduplicates_urls(self, sample_search_results_with_duplicates):
        """Test that duplicate URLs are deduplicated."""
        from local_deep_research.utilities.search_utilities import (
            extract_links_from_search_results,
            format_links_to_markdown,
        )

        links = extract_links_from_search_results(
            sample_search_results_with_duplicates
        )
        result = format_links_to_markdown(links)

        # Count URL occurrences
        url_count = result.count("https://example.com/article1")
        # Should appear only once (deduplicated)
        assert url_count == 1

    def test_groups_indices_for_same_url(self):
        """Test that indices are grouped for same URL."""
        from local_deep_research.utilities.search_utilities import (
            format_links_to_markdown,
        )

        links = [
            {"title": "Article", "url": "https://example.com", "index": "1"},
            {"title": "Article", "url": "https://example.com", "index": "3"},
        ]
        result = format_links_to_markdown(links)

        # Should have grouped indices [1, 3]
        assert "[1, 3]" in result

    def test_returns_empty_string_for_empty_list(self):
        """Test that empty list returns empty string."""
        from local_deep_research.utilities.search_utilities import (
            format_links_to_markdown,
        )

        result = format_links_to_markdown([])

        assert result == ""

    def test_handles_link_key_fallback(self):
        """Test that 'link' key is used as fallback for 'url'."""
        from local_deep_research.utilities.search_utilities import (
            format_links_to_markdown,
        )

        links = [{"title": "Test", "link": "https://example.com", "index": "1"}]
        result = format_links_to_markdown(links)

        assert "https://example.com" in result

    def test_uses_untitled_for_missing_title(self):
        """Test that 'Untitled' is used for missing title."""
        from local_deep_research.utilities.search_utilities import (
            format_links_to_markdown,
        )

        links = [{"url": "https://example.com", "index": "1"}]
        result = format_links_to_markdown(links)

        assert "Untitled" in result

    def test_includes_source_nr_fallback(self):
        """Test that source nr is included for accessibility."""
        from local_deep_research.utilities.search_utilities import (
            format_links_to_markdown,
        )

        links = [{"title": "Test", "url": "https://example.com", "index": "1"}]
        result = format_links_to_markdown(links)

        assert "source nr:" in result


class TestFormatFindings:
    """Tests for format_findings function."""

    def test_includes_synthesized_content(
        self, sample_findings, sample_questions_by_iteration
    ):
        """Test that synthesized content is included."""
        from local_deep_research.utilities.search_utilities import (
            format_findings,
        )

        result = format_findings(
            sample_findings,
            "This is the synthesis.",
            sample_questions_by_iteration,
        )

        assert "This is the synthesis." in result

    def test_includes_search_questions_section(
        self, sample_findings, sample_questions_by_iteration
    ):
        """Test that search questions section is included."""
        from local_deep_research.utilities.search_utilities import (
            format_findings,
        )

        result = format_findings(
            sample_findings, "Synthesis", sample_questions_by_iteration
        )

        assert "SEARCH QUESTIONS BY ITERATION" in result
        assert "What is the main topic?" in result

    def test_includes_detailed_findings_section(
        self, sample_findings, sample_questions_by_iteration
    ):
        """Test that detailed findings section is included."""
        from local_deep_research.utilities.search_utilities import (
            format_findings,
        )

        result = format_findings(
            sample_findings, "Synthesis", sample_questions_by_iteration
        )

        assert "DETAILED FINDINGS" in result

    def test_includes_all_sources_section(
        self, sample_findings, sample_questions_by_iteration
    ):
        """Test that all sources section is included."""
        from local_deep_research.utilities.search_utilities import (
            format_findings,
        )

        result = format_findings(
            sample_findings, "Synthesis", sample_questions_by_iteration
        )

        assert "ALL SOURCES" in result

    def test_formats_followup_phases(
        self, sample_findings, sample_questions_by_iteration
    ):
        """Test formatting of Follow-up phases."""
        from local_deep_research.utilities.search_utilities import (
            format_findings,
        )

        result = format_findings(
            sample_findings, "Synthesis", sample_questions_by_iteration
        )

        # Should include the follow-up phase header
        assert "Follow-up Iteration 1.1" in result

    def test_formats_subquery_phases(
        self, sample_findings_with_subquery, subquery_questions
    ):
        """Test formatting of Sub-query phases (IterDRAG)."""
        from local_deep_research.utilities.search_utilities import (
            format_findings,
        )

        result = format_findings(
            sample_findings_with_subquery, "Synthesis", subquery_questions
        )

        # Should include sub-query content
        assert "Content for sub-query 1" in result

    def test_handles_empty_findings(self, sample_questions_by_iteration):
        """Test handling of empty findings list."""
        from local_deep_research.utilities.search_utilities import (
            format_findings,
        )

        result = format_findings([], "Synthesis", sample_questions_by_iteration)

        # Should still include synthesis
        assert "Synthesis" in result
        # Should not have detailed findings section
        assert "DETAILED FINDINGS" not in result

    def test_handles_empty_questions(self, sample_findings):
        """Test handling of empty questions dictionary."""
        from local_deep_research.utilities.search_utilities import (
            format_findings,
        )

        result = format_findings(sample_findings, "Synthesis", {})

        # Should still include synthesis and findings
        assert "Synthesis" in result
        # Should not have questions section
        assert "SEARCH QUESTIONS BY ITERATION" not in result

    def test_extracts_sources_from_findings(
        self, sample_findings, sample_questions_by_iteration
    ):
        """Test that sources are extracted from findings."""
        from local_deep_research.utilities.search_utilities import (
            format_findings,
        )

        result = format_findings(
            sample_findings, "Synthesis", sample_questions_by_iteration
        )

        assert "source1.com" in result

    def test_handles_finding_with_question(self):
        """Test handling of finding with embedded question."""
        from local_deep_research.utilities.search_utilities import (
            format_findings,
        )

        findings = [
            {
                "phase": "Search",
                "content": "Content",
                "question": "What is the answer?",
                "search_results": [],
            }
        ]

        result = format_findings(findings, "Synthesis", {})

        assert "SEARCH QUESTION:" in result
        assert "What is the answer?" in result


class TestPrintSearchResults:
    """Tests for print_search_results function."""

    def test_logs_formatted_results(self, sample_search_results):
        """Test that results are logged."""
        from local_deep_research.utilities.search_utilities import (
            print_search_results,
        )

        with patch(
            "local_deep_research.utilities.search_utilities.logger"
        ) as mock_logger:
            print_search_results(sample_search_results)

            mock_logger.info.assert_called()

    def test_handles_empty_results(self):
        """Test handling of empty results."""
        from local_deep_research.utilities.search_utilities import (
            print_search_results,
        )

        with patch(
            "local_deep_research.utilities.search_utilities.logger"
        ) as mock_logger:
            print_search_results([])

            # Should still call logger (with empty string)
            mock_logger.info.assert_called()


class TestEdgeCases:
    """Tests for edge cases and error handling."""

    def test_extract_links_handles_exception(self):
        """Test that extract_links handles exceptions gracefully."""
        from local_deep_research.utilities.search_utilities import (
            extract_links_from_search_results,
        )

        # Create a result that could cause issues
        results = [
            {"title": "Valid", "link": "https://example.com"},
            object(),  # This would cause an error
            {"title": "Also Valid", "link": "https://test.com"},
        ]

        # Should not raise, should skip problematic items
        with patch("local_deep_research.utilities.search_utilities.logger"):
            links = extract_links_from_search_results(results)
            # Should get at least some valid results
            assert len(links) >= 0

    def test_format_findings_handles_none_search_results(self):
        """Test that format_findings handles None search_results."""
        from local_deep_research.utilities.search_utilities import (
            format_findings,
        )

        findings = [
            {
                "phase": "Test",
                "content": "Content",
                "search_results": None,  # None instead of list
            }
        ]

        # Should not raise
        result = format_findings(findings, "Synthesis", {})
        assert "Content" in result

    def test_format_findings_handles_malformed_phase(self):
        """Test that format_findings handles malformed phase strings."""
        from local_deep_research.utilities.search_utilities import (
            format_findings,
        )

        findings = [
            {
                "phase": "Follow-up Iteration X.Y",  # Malformed - not numbers
                "content": "Content",
                "search_results": [],
            }
        ]

        # Should not raise
        result = format_findings(findings, "Synthesis", {1: ["Question"]})
        assert "Content" in result



"""
Command-line interface for benchmarking.

This module provides CLI commands for running benchmarks.
"""

import argparse

# import logging - replaced with loguru
import sys
from loguru import logger

from ...config.paths import get_data_directory
from .. import (
    get_available_datasets,
    run_browsecomp_benchmark,
    run_simpleqa_benchmark,
)


def setup_benchmark_parser(subparsers):
    """
    Set up the benchmark CLI commands.

    Args:
        subparsers: argparse subparsers object to add commands to
    """
    # Common benchmark arguments
    benchmark_parent = argparse.ArgumentParser(add_help=False)
    benchmark_parent.add_argument(
        "--examples",
        type=int,
        default=100,
        help="Number of examples to run (default: 100)",
    )
    benchmark_parent.add_argument(
        "--iterations",
        type=int,
        default=3,
        help="Number of search iterations (default: 3)",
    )
    benchmark_parent.add_argument(
        "--questions",
        type=int,
        default=3,
        help="Questions per iteration (default: 3)",
    )
    benchmark_parent.add_argument(
        "--search-tool",
        type=str,
        default="searxng",
        help="Search tool to use (default: searxng)",
    )
    benchmark_parent.add_argument(
        "--output-dir",
        type=str,
        default=str(get_data_directory() / "benchmark_results"),
        help="Directory to save results (default: user data directory/benchmark_results)",
    )
    benchmark_parent.add_argument(
        "--human-eval",
        action="store_true",
        help="Use human evaluation instead of automated",
    )
    benchmark_parent.add_argument(
        "--eval-model", type=str, help="Model to use for evaluation"
    )
    benchmark_parent.add_argument(
        "--eval-provider", type=str, help="Provider to use for evaluation"
    )
    benchmark_parent.add_argument(
        "--custom-dataset", type=str, help="Path to custom dataset"
    )
    benchmark_parent.add_argument(
        "--no-eval", action="store_true", help="Skip evaluation phase"
    )

    # Add model configuration options for the search system
    benchmark_parent.add_argument(
        "--search-model", type=str, help="Model to use for the search system"
    )
    benchmark_parent.add_argument(
        "--search-provider",
        type=str,
        help="Provider to use for the search system",
    )
    benchmark_parent.add_argument(
        "--endpoint-url",
        type=str,
        help="Endpoint URL for OpenRouter or other API services",
    )
    benchmark_parent.add_argument(
        "--search-strategy",
        type=str,
        default="source_based",
        choices=["source_based", "standard", "rapid", "parallel", "iterdrag"],
        help="Search strategy to use (default: source_based)",
    )

    # SimpleQA benchmark command
    simpleqa_parser = subparsers.add_parser(
        "simpleqa", parents=[benchmark_parent], help="Run SimpleQA benchmark"
    )
    simpleqa_parser.set_defaults(func=run_simpleqa_cli)

    # BrowseComp benchmark command
    browsecomp_parser = subparsers.add_parser(
        "browsecomp",
        parents=[benchmark_parent],
        help="Run BrowseComp benchmark",
    )
    browsecomp_parser.set_defaults(func=run_browsecomp_cli)

    # List available benchmarks command
    list_parser = subparsers.add_parser(
        "list", help="List available benchmarks"
    )
    list_parser.set_defaults(func=list_benchmarks_cli)

    # Compare configurations command
    compare_parser = subparsers.add_parser(
        "compare", help="Compare multiple search configurations"
    )
    compare_parser.add_argument(
        "--dataset",
        type=str,
        default="simpleqa",
        choices=["simpleqa", "browsecomp"],
        help="Dataset to use for comparison",
    )
    compare_parser.add_argument(
        "--examples",
        type=int,
        default=20,
        help="Number of examples for each configuration (default: 20)",
    )
    compare_parser.add_argument(
        "--output-dir",
        type=str,
        default=str(get_data_directory() / "benchmark_results" / "comparison"),
        help="Directory to save comparison results",
    )
    compare_parser.set_defaults(func=compare_configs_cli)


def run_simpleqa_cli(args):
    """
    CLI handler for SimpleQA benchmark.

    Args:
        args: Parsed command-line arguments
    """
    # Set up search configuration
    search_config = {
        "iterations": args.iterations,
        "questions_per_iteration": args.questions,
        "search_tool": args.search_tool,
    }

    # Add model configuration if provided
    if hasattr(args, "search_model") and args.search_model:
        search_config["model_name"] = args.search_model
    if hasattr(args, "search_provider") and args.search_provider:
        search_config["provider"] = args.search_provider
    if hasattr(args, "endpoint_url") and args.endpoint_url:
        search_config["openai_endpoint_url"] = args.endpoint_url
    if hasattr(args, "search_strategy") and args.search_strategy:
        search_config["search_strategy"] = args.search_strategy

    # Set up evaluation configuration if needed
    evaluation_config = None
    if args.eval_model or args.eval_provider:
        evaluation_config = {}
        if args.eval_model:
            evaluation_config["model_name"] = args.eval_model
        if args.eval_provider:
            evaluation_config["provider"] = args.eval_provider

    # Run the benchmark
    result = run_simpleqa_benchmark(
        num_examples=args.examples,
        dataset_path=args.custom_dataset,
        output_dir=args.output_dir,
        search_config=search_config,
        evaluation_config=evaluation_config,
        human_evaluation=args.human_eval,
        run_evaluation=not args.no_eval,
    )

    # Print results summary
    if "metrics" in result:
        print("\nSimpleQA Benchmark Results:")
        print(f"  Accuracy: {result['metrics'].get('accuracy', 0):.3f}")
        print(f"  Total examples: {result['total_examples']}")
        print(f"  Correct answers: {result['metrics'].get('correct', 0)}")
        print(
            f"  Average time: {result['metrics'].get('average_processing_time', 0):.2f}s"
        )
        print(f"\nReport saved to: {result.get('report_path', 'N/A')}")
    else:
        print("\nSimpleQA Benchmark Completed (no evaluation)")
        print(f"  Total examples: {result['total_examples']}")
        print(f"  Results saved to: {result.get('results_path', 'N/A')}")


def run_browsecomp_cli(args):
    """
    CLI handler for BrowseComp benchmark.

    Args:
        args: Parsed command-line arguments
    """
    # Set up search configuration
    search_config = {
        "iterations": args.iterations,
        "questions_per_iteration": args.questions,
        "search_tool": args.search_tool,
    }

    # Add model configuration if provided
    if hasattr(args, "search_model") and args.search_model:
        search_config["model_name"] = args.search_model
    if hasattr(args, "search_provider") and args.search_provider:
        search_config["provider"] = args.search_provider
    if hasattr(args, "endpoint_url") and args.endpoint_url:
        search_config["openai_endpoint_url"] = args.endpoint_url
    if hasattr(args, "search_strategy") and args.search_strategy:
        search_config["search_strategy"] = args.search_strategy

    # Set up evaluation configuration if needed
    evaluation_config = None
    if args.eval_model or args.eval_provider:
        evaluation_config = {}
        if args.eval_model:
            evaluation_config["model_name"] = args.eval_model
        if args.eval_provider:
            evaluation_config["provider"] = args.eval_provider

    # Run the benchmark
    result = run_browsecomp_benchmark(
        num_examples=args.examples,
        dataset_path=args.custom_dataset,
        output_dir=args.output_dir,
        search_config=search_config,
        evaluation_config=evaluation_config,
        human_evaluation=args.human_eval,
        run_evaluation=not args.no_eval,
    )

    # Print results summary
    if "metrics" in result:
        print("\nBrowseComp Benchmark Results:")
        print(f"  Accuracy: {result['metrics'].get('accuracy', 0):.3f}")
        print(f"  Total examples: {result['total_examples']}")
        print(f"  Correct answers: {result['metrics'].get('correct', 0)}")
        print(
            f"  Average time: {result['metrics'].get('average_processing_time', 0):.2f}s"
        )
        print(f"\nReport saved to: {result.get('report_path', 'N/A')}")
    else:
        print("\nBrowseComp Benchmark Completed (no evaluation)")
        print(f"  Total examples: {result['total_examples']}")
        print(f"  Results saved to: {result.get('results_path', 'N/A')}")


def list_benchmarks_cli(args):
    """
    CLI handler for listing available benchmarks.

    Args:
        args: Parsed command-line arguments
    """
    datasets = get_available_datasets()

    print("\nAvailable Benchmarks:")
    for dataset in datasets:
        print(f"  {dataset['id']}: {dataset['name']}")
        print(f"    {dataset['description']}")
        print(f"    URL: {dataset['url']}")
        print()


def compare_configs_cli(args):
    """
    CLI handler for comparing multiple configurations.

    Args:
        args: Parsed command-line arguments
    """
    # Import the compare configurations function
    from ...api.benchmark_functions import compare_configurations

    # Run the comparison
    result = compare_configurations(
        dataset_type=args.dataset,
        num_examples=args.examples,
        output_dir=args.output_dir,
    )

    # Print results summary
    print("\nConfiguration Comparison Results:")
    print(f"  Dataset: {args.dataset}")
    print(f"  Configurations tested: {result['configurations_tested']}")
    print(f"  Report saved to: {result['report_path']}")

    # Print brief comparison table
    print("\nResults Summary:")
    print("Configuration   | Accuracy | Avg. Time")
    print("--------------- | -------- | ---------")
    for res in result["results"]:
        name = res["configuration_name"]
        acc = res.get("metrics", {}).get("accuracy", 0)
        time = res.get("metrics", {}).get("average_processing_time", 0)
        print(f"{name:15} | {acc:.3f}    | {time:.2f}s")


def main():
    """
    Main entry point for benchmark CLI.
    """
    parser = argparse.ArgumentParser(
        description="Local Deep Research Benchmarking Tool",
        prog="ldr-benchmark",
    )

    # Set up logging
    parser.add_argument(
        "--verbose", action="store_true", help="Enable verbose logging"
    )

    # Create subparsers
    subparsers = parser.add_subparsers(
        dest="command", help="Command to run", required=True
    )

    # Set up commands
    setup_benchmark_parser(subparsers)

    # Parse arguments
    args = parser.parse_args()

    # Set up logging with loguru
    if args.verbose:
        logger.add(sys.stderr, level="DEBUG")
    else:
        logger.add(sys.stderr, level="INFO")

    # Run command
    if hasattr(args, "func"):
        args.func(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



"""Tests for search_config module."""

from unittest.mock import MagicMock, patch


from local_deep_research.config.search_config import (
    get_search_snippets_only_setting,
    get_search,
    QUALITY_CHECK_DDG_URLS,
)


class TestQualityCheckConstant:
    """Tests for QUALITY_CHECK_DDG_URLS constant."""

    def test_is_boolean(self):
        """Should be a boolean value."""
        assert isinstance(QUALITY_CHECK_DDG_URLS, bool)


class TestGetSearchSnippetsOnlySetting:
    """Tests for get_search_snippets_only_setting function."""

    def test_returns_value_from_snapshot(self):
        """Should return value from settings snapshot."""
        snapshot = {"search.snippets_only": False}
        with patch(
            "local_deep_research.config.search_config.get_setting_from_snapshot",
            return_value=False,
        ):
            result = get_search_snippets_only_setting(
                settings_snapshot=snapshot
            )
            assert result is False

    def test_uses_default_true(self):
        """Should default to True."""
        with patch(
            "local_deep_research.config.search_config.get_setting_from_snapshot",
            return_value=True,
        ) as mock_get:
            get_search_snippets_only_setting()
            mock_get.assert_called_once()
            call_args = mock_get.call_args
            assert call_args[0][1] is True  # default argument


class TestGetSearch:
    """Tests for get_search function."""

    @patch("local_deep_research.config.search_config.factory_get_search")
    @patch("local_deep_research.config.search_config.get_llm")
    @patch("local_deep_research.config.search_config.get_setting_from_snapshot")
    def test_creates_search_engine(
        self, mock_get_setting, mock_get_llm, mock_factory
    ):
        """Should create search engine via factory."""
        mock_get_setting.side_effect = lambda key, default, **kwargs: default
        mock_llm = MagicMock()
        mock_get_llm.return_value = mock_llm
        mock_engine = MagicMock()
        mock_factory.return_value = mock_engine

        result = get_search()

        assert result is mock_engine
        mock_factory.assert_called_once()

    @patch("local_deep_research.config.search_config.factory_get_search")
    @patch("local_deep_research.config.search_config.get_llm")
    @patch("local_deep_research.config.search_config.get_setting_from_snapshot")
    def test_uses_provided_search_tool(
        self, mock_get_setting, mock_get_llm, mock_factory
    ):
        """Should use provided search_tool argument."""
        mock_get_setting.side_effect = lambda key, default, **kwargs: default
        mock_get_llm.return_value = MagicMock()
        mock_factory.return_value = MagicMock()

        get_search(search_tool="duckduckgo")

        call_kwargs = mock_factory.call_args[1]
        assert call_kwargs["search_tool"] == "duckduckgo"

    @patch("local_deep_research.config.search_config.factory_get_search")
    @patch("local_deep_research.config.search_config.get_llm")
    @patch("local_deep_research.config.search_config.get_setting_from_snapshot")
    def test_uses_provided_llm_instance(
        self, mock_get_setting, mock_get_llm, mock_factory
    ):
        """Should use provided llm_instance instead of getting new one."""
        mock_get_setting.side_effect = lambda key, default, **kwargs: default
        mock_factory.return_value = MagicMock()
        custom_llm = MagicMock()

        get_search(llm_instance=custom_llm)

        # Should not call get_llm
        mock_get_llm.assert_not_called()
        call_kwargs = mock_factory.call_args[1]
        assert call_kwargs["llm_instance"] is custom_llm

    @patch("local_deep_research.config.search_config.factory_get_search")
    @patch("local_deep_research.config.search_config.get_llm")
    @patch("local_deep_research.config.search_config.get_setting_from_snapshot")
    def test_extracts_value_from_dict_tool(
        self, mock_get_setting, mock_get_llm, mock_factory
    ):
        """Should extract value when search.tool is a dict."""

        def get_setting_side_effect(key, default, **kwargs):
            if key == "search.tool":
                return {"value": "wikipedia"}
            return default

        mock_get_setting.side_effect = get_setting_side_effect
        mock_get_llm.return_value = MagicMock()
        mock_factory.return_value = MagicMock()

        get_search()

        call_kwargs = mock_factory.call_args[1]
        assert call_kwargs["search_tool"] == "wikipedia"

    @patch("local_deep_research.config.search_config.factory_get_search")
    @patch("local_deep_research.config.search_config.get_llm")
    @patch("local_deep_research.config.search_config.get_setting_from_snapshot")
    def test_adds_username_to_snapshot(
        self, mock_get_setting, mock_get_llm, mock_factory
    ):
        """Should add username to settings snapshot."""
        mock_get_setting.side_effect = lambda key, default, **kwargs: default
        mock_get_llm.return_value = MagicMock()
        mock_factory.return_value = MagicMock()

        get_search(username="testuser", settings_snapshot={"existing": "value"})

        call_kwargs = mock_factory.call_args[1]
        assert call_kwargs["settings_snapshot"]["_username"] == "testuser"
        assert call_kwargs["settings_snapshot"]["existing"] == "value"

    @patch("local_deep_research.config.search_config.factory_get_search")
    @patch("local_deep_research.config.search_config.get_llm")
    @patch("local_deep_research.config.search_config.get_setting_from_snapshot")
    def test_creates_snapshot_with_username_only(
        self, mock_get_setting, mock_get_llm, mock_factory
    ):
        """Should create snapshot with just username if none provided."""
        mock_get_setting.side_effect = lambda key, default, **kwargs: default
        mock_get_llm.return_value = MagicMock()
        mock_factory.return_value = MagicMock()

        get_search(username="testuser")

        call_kwargs = mock_factory.call_args[1]
        assert call_kwargs["settings_snapshot"]["_username"] == "testuser"

    @patch("local_deep_research.config.search_config.factory_get_search")
    @patch("local_deep_research.config.search_config.get_llm")
    @patch("local_deep_research.config.search_config.get_setting_from_snapshot")
    def test_passes_programmatic_mode(
        self, mock_get_setting, mock_get_llm, mock_factory
    ):
        """Should pass programmatic_mode to factory."""
        mock_get_setting.side_effect = lambda key, default, **kwargs: default
        mock_get_llm.return_value = MagicMock()
        mock_factory.return_value = MagicMock()

        get_search(programmatic_mode=True)

        call_kwargs = mock_factory.call_args[1]
        assert call_kwargs["programmatic_mode"] is True

    @patch("local_deep_research.config.search_config.factory_get_search")
    @patch("local_deep_research.config.search_config.get_llm")
    @patch("local_deep_research.config.search_config.get_setting_from_snapshot")
    def test_passes_all_search_params(
        self, mock_get_setting, mock_get_llm, mock_factory
    ):
        """Should pass all search parameters to factory."""
        settings = {
            "search.tool": "searxng",
            "search.max_results": 20,
            "search.region": "us-en",
            "search.time_period": "w",
            "search.safe_search": False,
            "search.snippets_only": False,
            "search.search_language": "German",
            "search.max_filtered_results": 10,
        }

        def get_setting_side_effect(key, default, **kwargs):
            return settings.get(key, default)

        mock_get_setting.side_effect = get_setting_side_effect
        mock_get_llm.return_value = MagicMock()
        mock_factory.return_value = MagicMock()

        get_search()

        call_kwargs = mock_factory.call_args[1]
        assert call_kwargs["max_results"] == 20
        assert call_kwargs["region"] == "us-en"
        assert call_kwargs["time_period"] == "w"
        assert call_kwargs["safe_search"] is False
        assert call_kwargs["search_snippets_only"] is False
        assert call_kwargs["search_language"] == "German"
        assert call_kwargs["max_filtered_results"] == 10

    @patch("local_deep_research.config.search_config.factory_get_search")
    @patch("local_deep_research.config.search_config.get_llm")
    @patch("local_deep_research.config.search_config.get_setting_from_snapshot")
    def test_returns_none_when_factory_fails(
        self, mock_get_setting, mock_get_llm, mock_factory
    ):
        """Should return None when factory returns None."""
        mock_get_setting.side_effect = lambda key, default, **kwargs: default
        mock_get_llm.return_value = MagicMock()
        mock_factory.return_value = None

        result = get_search()

        assert result is None



"""
Tests for SQLCipher missing scenario.

Verifies that users get helpful error messages when SQLCipher is not installed
and that the LDR_ALLOW_UNENCRYPTED workaround works.
"""

import os
import pytest
from unittest.mock import patch

from tests.test_utils import add_src_to_path

add_src_to_path()


class TestSQLCipherMissing:
    """Test behavior when SQLCipher is not available."""

    def test_error_message_mentions_sqlcipher(self):
        """Error message should mention SQLCipher so users know what's missing."""
        from src.local_deep_research.database.encrypted_db import (
            DatabaseManager,
        )

        old_value = os.environ.pop("LDR_ALLOW_UNENCRYPTED", None)

        try:
            with patch(
                "src.local_deep_research.database.encrypted_db.get_sqlcipher_module"
            ) as mock_get:
                mock_get.side_effect = ImportError(
                    "No module named 'sqlcipher3'"
                )

                with pytest.raises(RuntimeError) as exc_info:
                    DatabaseManager()

                error_msg = str(exc_info.value)
                assert "SQLCipher" in error_msg, (
                    f"Error should mention SQLCipher. Got: {error_msg}"
                )
        finally:
            if old_value is not None:
                os.environ["LDR_ALLOW_UNENCRYPTED"] = old_value

    def test_error_message_mentions_workaround(self):
        """Error message should mention LDR_ALLOW_UNENCRYPTED workaround."""
        from src.local_deep_research.database.encrypted_db import (
            DatabaseManager,
        )

        old_value = os.environ.pop("LDR_ALLOW_UNENCRYPTED", None)

        try:
            with patch(
                "src.local_deep_research.database.encrypted_db.get_sqlcipher_module"
            ) as mock_get:
                mock_get.side_effect = ImportError(
                    "No module named 'sqlcipher3'"
                )

                with pytest.raises(RuntimeError) as exc_info:
                    DatabaseManager()

                error_msg = str(exc_info.value)
                assert "LDR_ALLOW_UNENCRYPTED" in error_msg, (
                    f"Error should mention workaround. Got: {error_msg}"
                )
        finally:
            if old_value is not None:
                os.environ["LDR_ALLOW_UNENCRYPTED"] = old_value

    def test_workaround_allows_startup_without_encryption(self):
        """LDR_ALLOW_UNENCRYPTED=true should allow startup without SQLCipher."""
        from src.local_deep_research.database.encrypted_db import (
            DatabaseManager,
        )

        old_value = os.environ.get("LDR_ALLOW_UNENCRYPTED")
        os.environ["LDR_ALLOW_UNENCRYPTED"] = "true"

        try:
            with patch(
                "src.local_deep_research.database.encrypted_db.get_sqlcipher_module"
            ) as mock_get:
                mock_get.side_effect = ImportError(
                    "No module named 'sqlcipher3'"
                )

                # Should NOT raise
                manager = DatabaseManager()
                assert manager.has_encryption is False, (
                    "With workaround and no SQLCipher, has_encryption should be False"
                )
        finally:
            if old_value is not None:
                os.environ["LDR_ALLOW_UNENCRYPTED"] = old_value
            else:
                os.environ.pop("LDR_ALLOW_UNENCRYPTED", None)

    def test_db_manager_has_encryption_is_boolean(self):
        """db_manager.has_encryption should be a boolean."""
        from src.local_deep_research.database.encrypted_db import db_manager

        assert isinstance(db_manager.has_encryption, bool), (
            f"has_encryption should be bool, got {type(db_manager.has_encryption)}"
        )



"""
Test script to demonstrate the duplicate links issue and our fix for issue #301.

This test demonstrates the problem where the search system duplicates links in detailed report mode
by unconditionally extending the all_links_of_system list with itself.

GitHub issue: https://github.com/LearningCircuit/local-deep-research/issues/301

The fix is to check if the lists are the same object (have the same id()) before extending.
"""


class Strategy:
    def __init__(self, all_links=None):
        # Create a new list if None is provided
        self.all_links_of_system = [] if all_links is None else all_links
        print(
            f"Strategy initialized with list id: {id(self.all_links_of_system)}"
        )

    def analyze_topic(self, query):
        # Add some links
        self.all_links_of_system.extend(
            [
                {"title": "Link 1", "link": "http://example.com/1"},
                {"title": "Link 2", "link": "http://example.com/2"},
            ]
        )
        print(f"Strategy now has {len(self.all_links_of_system)} links")
        return {"content": "Analysis results"}


class AdvancedSearchSystem:
    def __init__(self):
        # Initialize with empty list
        self.all_links_of_system = []
        print(
            f"Search system initialized with list id: {id(self.all_links_of_system)}"
        )

        # Create strategy with our list reference
        self.strategy = Strategy(all_links=self.all_links_of_system)

        # Check if they're the same object
        print(
            f"Are lists the same object? {id(self.all_links_of_system) == id(self.strategy.all_links_of_system)}"
        )

    def analyze_topic_with_bug(self, query):
        # Run the strategy
        result = self.strategy.analyze_topic(query)

        # BUG: Unconditionally extend our list with the strategy's list
        # This is problematic because they're the same list
        print("\nBUG DEMO: Extending unconditionally")
        before_count = len(self.all_links_of_system)
        print(f"Before extending: {before_count} links")

        self.all_links_of_system.extend(self.strategy.all_links_of_system)

        after_count = len(self.all_links_of_system)
        print(f"After extending: {after_count} links")
        print(f"Added {after_count - before_count} links (duplicates)")

        # Return results
        return {
            "all_links_of_system": self.all_links_of_system,
            "content": result["content"],
        }

    def analyze_topic_with_fix(self, query):
        # Run the strategy
        result = self.strategy.analyze_topic(query)

        # FIX: Only extend if they're different objects
        print("\nFIX DEMO: Only extending if lists are different")
        before_count = len(self.all_links_of_system)
        print(f"Before fix check: {before_count} links")

        if id(self.all_links_of_system) != id(
            self.strategy.all_links_of_system
        ):
            print("Lists are different objects - extending")
            self.all_links_of_system.extend(self.strategy.all_links_of_system)
        else:
            print(
                "Lists are the same object - not extending (avoiding duplicates)"
            )

        after_count = len(self.all_links_of_system)
        print(f"After fix check: {after_count} links")
        print(f"Added {after_count - before_count} links")

        # Return results
        return {
            "all_links_of_system": self.all_links_of_system,
            "content": result["content"],
        }


def test_bug_and_fix():
    print("=== Testing Bug and Fix ===\n")

    print("1. Creating first search system to demonstrate the bug:")
    search_system1 = AdvancedSearchSystem()
    result1 = search_system1.analyze_topic_with_bug(
        "What is quantum computing?"
    )
    print(f"\nFinal link count with bug: {len(result1['all_links_of_system'])}")

    print("\n2. Creating second search system to demonstrate the fix:")
    search_system2 = AdvancedSearchSystem()
    result2 = search_system2.analyze_topic_with_fix(
        "What is quantum computing?"
    )
    print(f"\nFinal link count with fix: {len(result2['all_links_of_system'])}")

    print("\nTest complete!")


if __name__ == "__main__":
    test_bug_and_fix()



"""Queue manager for handling research queue operations"""

from loguru import logger
from sqlalchemy import func
from sqlalchemy.orm import sessionmaker

from ...database.encrypted_db import db_manager
from ...database.models import QueuedResearch, ResearchHistory
from .processor_v2 import queue_processor


class QueueManager:
    """Manages the research queue operations"""

    @staticmethod
    def add_to_queue(username, research_id, query, mode, settings):
        """
        Add a research to the queue

        Args:
            username: User who owns the research
            research_id: UUID of the research
            query: Research query
            mode: Research mode
            settings: Research settings dictionary

        Returns:
            int: Queue position
        """
        engine = db_manager.connections.get(username)
        if not engine:
            raise ValueError(f"No database connection for user {username}")

        SessionLocal = sessionmaker(bind=engine)
        session = SessionLocal()

        try:
            # Get the next position in queue for this user
            max_position = (
                session.query(func.max(QueuedResearch.position))
                .filter_by(username=username)
                .scalar()
                or 0
            )

            queued_record = QueuedResearch(
                username=username,
                research_id=research_id,
                query=query,
                mode=mode,
                settings_snapshot=settings,
                position=max_position + 1,
            )
            session.add(queued_record)
            session.commit()

            logger.info(
                f"Added research {research_id} to queue at position {max_position + 1}"
            )

            # Send RESEARCH_QUEUED notification if enabled
            try:
                from ...settings import SettingsManager
                from ...notifications import send_queue_notification

                settings_manager = SettingsManager(session)
                settings_snapshot = settings_manager.get_settings_snapshot()

                send_queue_notification(
                    username=username,
                    research_id=research_id,
                    query=query,
                    settings_snapshot=settings_snapshot,
                    position=max_position + 1,
                )
            except Exception as e:
                logger.debug(f"Failed to send queued notification: {e}")

            # Notify queue processor about the new queued research
            # Note: When using QueueManager, we don't have all parameters for direct execution
            # So it will fall back to queue mode
            queue_processor.notify_research_queued(username, research_id)

            return max_position + 1

        finally:
            session.close()

    @staticmethod
    def get_queue_position(username, research_id):
        """
        Get the current queue position for a research

        Args:
            username: User who owns the research
            research_id: UUID of the research

        Returns:
            int: Current queue position or None if not in queue
        """
        engine = db_manager.connections.get(username)
        if not engine:
            return None

        SessionLocal = sessionmaker(bind=engine)
        session = SessionLocal()

        try:
            queued = (
                session.query(QueuedResearch)
                .filter_by(username=username, research_id=research_id)
                .first()
            )

            if not queued:
                return None

            # Count how many are ahead in queue
            ahead_count = (
                session.query(QueuedResearch)
                .filter(
                    QueuedResearch.username == username,
                    QueuedResearch.position < queued.position,
                )
                .count()
            )

            return ahead_count + 1

        finally:
            session.close()

    @staticmethod
    def remove_from_queue(username, research_id):
        """
        Remove a research from the queue

        Args:
            username: User who owns the research
            research_id: UUID of the research

        Returns:
            bool: True if removed, False if not found
        """
        engine = db_manager.connections.get(username)
        if not engine:
            return False

        SessionLocal = sessionmaker(bind=engine)
        session = SessionLocal()

        try:
            queued = (
                session.query(QueuedResearch)
                .filter_by(username=username, research_id=research_id)
                .first()
            )

            if not queued:
                return False

            position = queued.position
            session.delete(queued)

            # Update positions of items behind in queue
            session.query(QueuedResearch).filter(
                QueuedResearch.username == username,
                QueuedResearch.position > position,
            ).update({QueuedResearch.position: QueuedResearch.position - 1})

            session.commit()
            logger.info(f"Removed research {research_id} from queue")
            return True

        finally:
            session.close()

    @staticmethod
    def get_user_queue(username):
        """
        Get all queued researches for a user

        Args:
            username: User to get queue for

        Returns:
            list: List of queued research info
        """
        engine = db_manager.connections.get(username)
        if not engine:
            return []

        SessionLocal = sessionmaker(bind=engine)
        session = SessionLocal()

        try:
            queued_items = (
                session.query(QueuedResearch)
                .filter_by(username=username)
                .order_by(QueuedResearch.position)
                .all()
            )

            result = []
            for item in queued_items:
                # Get research info
                research = (
                    session.query(ResearchHistory)
                    .filter_by(id=item.research_id)
                    .first()
                )

                if research:
                    result.append(
                        {
                            "research_id": item.research_id,
                            "query": item.query,
                            "mode": item.mode,
                            "position": item.position,
                            "created_at": item.created_at.isoformat()
                            if item.created_at
                            else None,
                            "is_processing": item.is_processing,
                        }
                    )

            return result

        finally:
            session.close()



"""Tests for log_utils module."""

import logging
import queue
import threading
from datetime import datetime
from unittest.mock import Mock, patch, MagicMock


class TestInterceptHandler:
    """Tests for InterceptHandler class."""

    def test_emit_forwards_to_loguru(self):
        """Should forward log records to loguru."""
        from local_deep_research.utilities.log_utils import InterceptHandler

        handler = InterceptHandler()

        # Create a mock log record
        record = logging.LogRecord(
            name="test",
            level=logging.INFO,
            pathname="test.py",
            lineno=1,
            msg="Test message",
            args=(),
            exc_info=None,
        )

        with patch(
            "local_deep_research.utilities.log_utils.logger"
        ) as mock_logger:
            mock_logger.level.return_value = Mock(name="INFO")
            mock_opt = Mock()
            mock_logger.opt.return_value = mock_opt

            handler.emit(record)

            mock_logger.opt.assert_called()
            mock_opt.log.assert_called()

    def test_handles_unknown_level(self):
        """Should handle unknown log levels by using levelno."""
        from local_deep_research.utilities.log_utils import InterceptHandler

        handler = InterceptHandler()

        record = logging.LogRecord(
            name="test",
            level=35,  # Non-standard level
            pathname="test.py",
            lineno=1,
            msg="Test",
            args=(),
            exc_info=None,
        )
        record.levelname = "CUSTOM"

        with patch(
            "local_deep_research.utilities.log_utils.logger"
        ) as mock_logger:
            mock_logger.level.side_effect = ValueError("Unknown level")
            mock_opt = Mock()
            mock_logger.opt.return_value = mock_opt

            handler.emit(record)

            mock_opt.log.assert_called()


class TestLogForResearch:
    """Tests for log_for_research decorator."""

    def test_sets_research_id_in_g(self):
        """Should set research_id in Flask g object."""
        from local_deep_research.utilities.log_utils import log_for_research

        mock_g = MagicMock()

        @log_for_research
        def test_func(research_id):
            return "done"

        with patch("local_deep_research.utilities.log_utils.g", mock_g):
            test_func("test-uuid-123")

            # Check that research_id was set
            assert mock_g.research_id == "test-uuid-123"

    def test_removes_research_id_after_function(self):
        """Should remove research_id from g after function completes."""
        from local_deep_research.utilities.log_utils import log_for_research

        @log_for_research
        def test_func(research_id):
            return "result"

        mock_g = MagicMock()
        with patch("local_deep_research.utilities.log_utils.g", mock_g):
            test_func("uuid")

            mock_g.pop.assert_called_with("research_id")

    def test_preserves_function_metadata(self):
        """Should preserve function name and docstring."""
        from local_deep_research.utilities.log_utils import log_for_research

        @log_for_research
        def documented_func(research_id):
            """My documentation."""
            pass

        assert documented_func.__name__ == "documented_func"
        assert documented_func.__doc__ == "My documentation."

    def test_passes_args_and_kwargs(self):
        """Should pass arguments correctly."""
        from local_deep_research.utilities.log_utils import log_for_research

        @log_for_research
        def test_func(research_id, arg1, kwarg1=None):
            return (research_id, arg1, kwarg1)

        mock_g = MagicMock()
        with patch("local_deep_research.utilities.log_utils.g", mock_g):
            result = test_func("uuid", "value1", kwarg1="value2")

            assert result == ("uuid", "value1", "value2")


class TestDatabaseSink:
    """Tests for database_sink function."""

    def test_creates_log_entry_dict(self):
        """Should create log entry dictionary from message."""
        from local_deep_research.utilities.log_utils import database_sink

        mock_message = Mock()
        mock_message.record = {
            "time": datetime.now(),
            "message": "Test log message",
            "name": "test_module",
            "function": "test_function",
            "line": 42,
            "level": Mock(name="INFO"),
            "extra": {},
        }

        with patch(
            "local_deep_research.utilities.log_utils.has_app_context",
            return_value=False,
        ):
            with patch(
                "local_deep_research.utilities.log_utils._log_queue"
            ) as mock_queue:
                database_sink(mock_message)

                # Should queue the log since we're not in app context
                mock_queue.put_nowait.assert_called_once()

    def test_queues_log_from_non_main_thread(self):
        """Should queue log when not in main thread."""
        from local_deep_research.utilities.log_utils import database_sink
        import local_deep_research.utilities.log_utils as module

        mock_message = Mock()
        mock_message.record = {
            "time": datetime.now(),
            "message": "Test message",
            "name": "module",
            "function": "func",
            "line": 1,
            "level": Mock(name="DEBUG"),
            "extra": {"research_id": "test-uuid"},
        }

        # Mock has_app_context to return True but thread name is not MainThread
        mock_thread = Mock()
        mock_thread.name = "WorkerThread"

        with patch.object(module, "has_app_context", return_value=True):
            with patch.object(
                module, "_get_research_id", return_value="test-uuid"
            ):
                with patch.object(
                    threading, "current_thread", return_value=mock_thread
                ):
                    with patch.object(module, "_log_queue") as mock_queue:
                        database_sink(mock_message)

                        # Should queue since not MainThread
                        mock_queue.put_nowait.assert_called_once()

    def test_handles_full_queue_gracefully(self):
        """Should not raise when queue is full."""
        from local_deep_research.utilities.log_utils import database_sink

        mock_message = Mock()
        mock_message.record = {
            "time": datetime.now(),
            "message": "Test",
            "name": "mod",
            "function": "f",
            "line": 1,
            "level": Mock(name="INFO"),
            "extra": {},
        }

        with patch(
            "local_deep_research.utilities.log_utils.has_app_context",
            return_value=False,
        ):
            with patch(
                "local_deep_research.utilities.log_utils._log_queue"
            ) as mock_queue:
                mock_queue.put_nowait.side_effect = queue.Full()

                # Should not raise
                database_sink(mock_message)

    def test_writes_to_database_in_main_thread(self):
        """Should write to database when in main thread with app context."""
        from local_deep_research.utilities.log_utils import database_sink
        import local_deep_research.utilities.log_utils as module

        mock_message = Mock()
        mock_message.record = {
            "time": datetime.now(),
            "message": "Test message",
            "name": "test_module",
            "function": "test_func",
            "line": 42,
            "level": Mock(name="INFO"),
            "extra": {"username": "testuser", "research_id": "test-uuid"},
        }

        mock_thread = Mock()
        mock_thread.name = "MainThread"

        mock_session = MagicMock()
        mock_cm = MagicMock()
        mock_cm.__enter__ = Mock(return_value=mock_session)
        mock_cm.__exit__ = Mock(return_value=None)

        mock_g = Mock()
        mock_g.get.return_value = None

        with patch.object(module, "has_app_context", return_value=True):
            with patch.object(module, "g", mock_g):
                with patch.object(
                    threading, "current_thread", return_value=mock_thread
                ):
                    with patch(
                        "local_deep_research.database.session_context.get_user_db_session",
                        return_value=mock_cm,
                    ):
                        database_sink(mock_message)

                        # Should write to database
                        mock_session.add.assert_called_once()
                        mock_session.commit.assert_called_once()

    def test_handles_database_error_gracefully(self):
        """Should not raise on database errors when writing."""
        from local_deep_research.utilities.log_utils import database_sink
        import local_deep_research.utilities.log_utils as module

        mock_message = Mock()
        mock_message.record = {
            "time": datetime.now(),
            "message": "Test",
            "name": "mod",
            "function": "f",
            "line": 1,
            "level": Mock(name="INFO"),
            "extra": {"research_id": "test-uuid"},
        }

        mock_thread = Mock()
        mock_thread.name = "MainThread"

        mock_g = Mock()
        mock_g.get.return_value = None

        with patch.object(module, "has_app_context", return_value=True):
            with patch.object(module, "g", mock_g):
                with patch.object(
                    threading, "current_thread", return_value=mock_thread
                ):
                    with patch(
                        "local_deep_research.database.session_context.get_user_db_session",
                        side_effect=Exception("DB error"),
                    ):
                        # Should not raise
                        database_sink(mock_message)

    def test_extracts_research_id_from_record_extra(self):
        """Should extract research_id from record extra."""
        from local_deep_research.utilities.log_utils import database_sink
        import local_deep_research.utilities.log_utils as module

        mock_message = Mock()
        mock_message.record = {
            "time": datetime.now(),
            "message": "Test",
            "name": "mod",
            "function": "f",
            "line": 1,
            "level": Mock(name="INFO"),
            "extra": {"research_id": "record-uuid"},
        }

        with patch.object(module, "has_app_context", return_value=False):
            with patch.object(module, "_log_queue") as mock_queue:
                database_sink(mock_message)

                # Verify the queued log entry contains the research_id
                call_args = mock_queue.put_nowait.call_args[0][0]
                assert call_args["research_id"] == "record-uuid"

    def test_extracts_research_id_from_flask_g(self):
        """Should extract research_id from Flask g when not in record."""
        from local_deep_research.utilities.log_utils import database_sink
        import local_deep_research.utilities.log_utils as module

        mock_message = Mock()
        mock_message.record = {
            "time": datetime.now(),
            "message": "Test",
            "name": "mod",
            "function": "f",
            "line": 1,
            "level": Mock(name="INFO"),
            "extra": {},
        }

        mock_g = Mock()
        mock_g.get.return_value = "flask-uuid"

        with patch.object(module, "has_app_context", return_value=True):
            with patch.object(module, "g", mock_g):
                with patch.object(module, "_log_queue") as mock_queue:
                    # Use non-main thread to queue instead of write
                    mock_thread = Mock()
                    mock_thread.name = "WorkerThread"
                    with patch.object(
                        threading, "current_thread", return_value=mock_thread
                    ):
                        database_sink(mock_message)

                        # Verify the queued log entry contains the research_id
                        call_args = mock_queue.put_nowait.call_args[0][0]
                        assert call_args["research_id"] == "flask-uuid"

    def test_record_research_id_takes_priority_over_flask(self):
        """Record research_id should take priority over Flask g."""
        from local_deep_research.utilities.log_utils import database_sink
        import local_deep_research.utilities.log_utils as module

        mock_message = Mock()
        mock_message.record = {
            "time": datetime.now(),
            "message": "Test",
            "name": "mod",
            "function": "f",
            "line": 1,
            "level": Mock(name="INFO"),
            "extra": {"research_id": "record-uuid"},
        }

        mock_g = Mock()
        mock_g.get.return_value = "flask-uuid"

        with patch.object(module, "has_app_context", return_value=True):
            with patch.object(module, "g", mock_g):
                with patch.object(module, "_log_queue") as mock_queue:
                    # Use non-main thread to queue instead of write
                    mock_thread = Mock()
                    mock_thread.name = "WorkerThread"
                    with patch.object(
                        threading, "current_thread", return_value=mock_thread
                    ):
                        database_sink(mock_message)

                        # Record research_id should win
                        call_args = mock_queue.put_nowait.call_args[0][0]
                        assert call_args["research_id"] == "record-uuid"


class TestFrontendProgressSink:
    """Tests for frontend_progress_sink function."""

    def test_skips_when_no_research_id(self):
        """Should skip when no research_id is available."""
        from local_deep_research.utilities.log_utils import (
            frontend_progress_sink,
        )

        mock_message = Mock()
        mock_message.record = {
            "message": "Test",
            "level": Mock(name="INFO"),
            "time": Mock(isoformat=Mock(return_value="2024-01-01T00:00:00")),
            "extra": {},
        }

        with patch(
            "local_deep_research.utilities.log_utils._get_research_id",
            return_value=None,
        ):
            with patch(
                "local_deep_research.utilities.log_utils.SocketIOService"
            ) as mock_socket:
                frontend_progress_sink(mock_message)

                # Should not emit anything
                mock_socket.return_value.emit_to_subscribers.assert_not_called()

    def test_emits_to_subscribers_with_research_id(self):
        """Should emit to subscribers when research_id is present."""
        from local_deep_research.utilities.log_utils import (
            frontend_progress_sink,
        )

        mock_message = Mock()
        mock_message.record = {
            "message": "Progress update",
            "level": Mock(name="INFO"),
            "time": Mock(isoformat=Mock(return_value="2024-01-01T12:00:00")),
            "extra": {"research_id": "test-uuid"},
        }

        with patch(
            "local_deep_research.utilities.log_utils._get_research_id",
            return_value="test-uuid",
        ):
            with patch(
                "local_deep_research.utilities.log_utils.SocketIOService"
            ) as mock_socket:
                frontend_progress_sink(mock_message)

                mock_socket.return_value.emit_to_subscribers.assert_called_once()
                call_args = (
                    mock_socket.return_value.emit_to_subscribers.call_args
                )
                assert call_args[0][0] == "progress"
                assert call_args[0][1] == "test-uuid"


class TestFlushLogQueue:
    """Tests for flush_log_queue function."""

    def test_flushes_all_queued_logs(self):
        """Should flush all logs from queue."""
        from local_deep_research.utilities.log_utils import flush_log_queue
        import local_deep_research.utilities.log_utils as module

        log_entries = [
            {
                "timestamp": datetime.now(),
                "message": "Log 1",
                "module": "mod",
                "function": "f",
                "line_no": 1,
                "level": "INFO",
                "research_id": None,
                "username": None,
            },
            {
                "timestamp": datetime.now(),
                "message": "Log 2",
                "module": "mod",
                "function": "f",
                "line_no": 2,
                "level": "INFO",
                "research_id": None,
                "username": None,
            },
        ]

        mock_session = MagicMock()
        mock_cm = MagicMock()
        mock_cm.__enter__ = Mock(return_value=mock_session)
        mock_cm.__exit__ = Mock(return_value=None)

        with patch.object(module, "_log_queue") as mock_queue:
            mock_queue.empty.side_effect = [False, False, True]
            mock_queue.get_nowait.side_effect = log_entries + [queue.Empty()]

            with patch(
                "local_deep_research.database.session_context.get_user_db_session",
                return_value=mock_cm,
            ):
                flush_log_queue()

                # Should have written 2 logs
                assert mock_session.add.call_count == 2
                assert mock_session.commit.call_count == 2

    def test_handles_empty_queue(self):
        """Should handle empty queue gracefully."""
        from local_deep_research.utilities.log_utils import flush_log_queue

        with patch(
            "local_deep_research.utilities.log_utils._log_queue"
        ) as mock_queue:
            mock_queue.empty.return_value = True

            # Should not raise
            flush_log_queue()


class TestConfigLogger:
    """Tests for config_logger function."""

    def test_configures_logger(self):
        """Should configure logger with sinks."""
        from local_deep_research.utilities.log_utils import config_logger

        with patch(
            "local_deep_research.utilities.log_utils.logger"
        ) as mock_logger:
            config_logger("test_app")

            mock_logger.enable.assert_called_with("local_deep_research")
            mock_logger.remove.assert_called_once()
            # Should add multiple sinks
            assert mock_logger.add.call_count >= 3

    def test_adds_file_logging_when_enabled(self):
        """Should add file logging when environment variable is set."""
        from local_deep_research.utilities.log_utils import config_logger

        with patch.dict("os.environ", {"LDR_ENABLE_FILE_LOGGING": "true"}):
            with patch(
                "local_deep_research.utilities.log_utils.logger"
            ) as mock_logger:
                config_logger("test_app")

                # Should add 4 sinks (stderr, database, frontend, file)
                assert mock_logger.add.call_count >= 4

    def test_creates_milestone_level(self):
        """Should create MILESTONE log level."""
        from local_deep_research.utilities.log_utils import config_logger

        with patch(
            "local_deep_research.utilities.log_utils.logger"
        ) as mock_logger:
            config_logger("test_app")

            mock_logger.level.assert_called()

    def test_handles_existing_milestone_level(self):
        """Should handle case where MILESTONE level already exists."""
        from local_deep_research.utilities.log_utils import config_logger

        with patch(
            "local_deep_research.utilities.log_utils.logger"
        ) as mock_logger:
            mock_logger.level.side_effect = ValueError("Level already exists")

            # Should not raise
            config_logger("test_app")


class TestWriteLogToDatabase:
    """Tests for _write_log_to_database function."""

    def test_writes_research_log(self):
        """Should write ResearchLog to database."""
        from local_deep_research.utilities.log_utils import (
            _write_log_to_database,
        )

        log_entry = {
            "timestamp": datetime.now(),
            "message": "Test message",
            "module": "test_module",
            "function": "test_func",
            "line_no": 42,
            "level": "INFO",
            "research_id": "test-uuid",
            "username": "testuser",
        }

        mock_session = MagicMock()
        mock_cm = MagicMock()
        mock_cm.__enter__ = Mock(return_value=mock_session)
        mock_cm.__exit__ = Mock(return_value=None)

        with patch(
            "local_deep_research.database.session_context.get_user_db_session",
            return_value=mock_cm,
        ):
            _write_log_to_database(log_entry)

            mock_session.add.assert_called_once()
            mock_session.commit.assert_called_once()

    def test_handles_database_error_gracefully(self):
        """Should not raise on database errors."""
        from local_deep_research.utilities.log_utils import (
            _write_log_to_database,
        )

        log_entry = {
            "timestamp": datetime.now(),
            "message": "Test",
            "module": "mod",
            "function": "f",
            "line_no": 1,
            "level": "INFO",
            "research_id": None,
            "username": None,
        }

        with patch(
            "local_deep_research.database.session_context.get_user_db_session",
            side_effect=Exception("DB error"),
        ):
            # Should not raise
            _write_log_to_database(log_entry)



"""
Fixtures for embedding provider tests.
"""

import pytest
from unittest.mock import Mock


@pytest.fixture
def mock_settings_snapshot():
    """Create a mock settings snapshot for testing."""
    return {
        "embeddings.ollama.model": "nomic-embed-text",
        "llm.ollama.url": "http://localhost:11434",
        "embeddings.openai.api_key": "test-openai-key",
        "embeddings.openai.model": "text-embedding-3-small",
    }


@pytest.fixture
def mock_empty_settings():
    """Create empty settings snapshot (no API keys configured)."""
    return {}


@pytest.fixture
def mock_ollama_models_response():
    """Create a mock Ollama models list response."""
    return {
        "models": [
            {"name": "nomic-embed-text:latest", "size": 274124704},
            {"name": "all-minilm:latest", "size": 45418096},
            {"name": "mxbai-embed-large:latest", "size": 669682064},
        ]
    }


@pytest.fixture
def mock_embedding_vectors():
    """Create mock embedding vectors for testing."""
    return [
        [0.1, 0.2, 0.3, 0.4, 0.5] * 76,  # 380-dimensional embedding
        [0.2, 0.3, 0.4, 0.5, 0.6] * 76,
    ]


@pytest.fixture
def mock_langchain_embeddings():
    """Create a mock LangChain Embeddings instance."""
    embeddings = Mock()
    embeddings.embed_documents.return_value = [
        [0.1, 0.2, 0.3] * 128,
        [0.2, 0.3, 0.4] * 128,
    ]
    embeddings.embed_query.return_value = [0.1, 0.2, 0.3] * 128
    return embeddings



"""
Comprehensive tests for PrecisionExtractionHandler.
Tests question type detection, various extraction methods, and edge cases.
"""

import pytest
from unittest.mock import Mock


class TestPrecisionExtractionHandlerInit:
    """Tests for PrecisionExtractionHandler initialization."""

    def test_init_with_llm_only(self, mock_llm):
        """Test initialization with just an LLM."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        handler = PrecisionExtractionHandler(llm=mock_llm)

        assert handler.llm == mock_llm
        assert handler.settings_snapshot == {}

    def test_init_creates_answer_patterns(self, mock_llm):
        """Test initialization creates regex patterns."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        handler = PrecisionExtractionHandler(llm=mock_llm)

        assert "full_name" in handler.answer_patterns
        assert "year" in handler.answer_patterns
        assert "number" in handler.answer_patterns
        assert "dimension" in handler.answer_patterns
        assert "score" in handler.answer_patterns

    def test_init_with_settings_snapshot(
        self, mock_llm, settings_with_fact_checking
    ):
        """Test initialization with settings snapshot."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        handler = PrecisionExtractionHandler(
            llm=mock_llm, settings_snapshot=settings_with_fact_checking
        )

        assert handler.settings_snapshot == settings_with_fact_checking


class TestPrecisionExtractionIdentifyQuestionType:
    """Tests for _identify_question_type method."""

    @pytest.fixture
    def handler(self, mock_llm):
        """Create a PrecisionExtractionHandler."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        return PrecisionExtractionHandler(llm=mock_llm)

    def test_identifies_full_name_question(self, handler):
        """Test detection of full name questions."""
        assert (
            handler._identify_question_type(
                "What is the full name of the founder?"
            )
            == "full_name"
        )

    def test_identifies_name_question_with_name_of(self, handler):
        """Test detection of name questions with 'name of'."""
        assert (
            handler._identify_question_type("What is the name of the CEO?")
            == "name"
        )

    def test_identifies_name_question_with_who_was(self, handler):
        """Test detection of name questions with 'who was'."""
        assert (
            handler._identify_question_type("Who was the inventor?") == "name"
        )

    def test_identifies_name_question_with_who_is(self, handler):
        """Test detection of name questions with 'who is'."""
        assert (
            handler._identify_question_type("Who is the current president?")
            == "name"
        )

    def test_identifies_location_question_with_where(self, handler):
        """Test detection of location questions with 'where'."""
        assert (
            handler._identify_question_type("Where did it happen?")
            == "location"
        )

    def test_identifies_location_question_with_city(self, handler):
        """Test detection of location questions mentioning city."""
        assert (
            handler._identify_question_type(
                "In which city was the company founded?"
            )
            == "location"
        )

    def test_identifies_location_question_with_country(self, handler):
        """Test detection of location questions mentioning country."""
        assert (
            handler._identify_question_type("What country is it located in?")
            == "location"
        )

    def test_identifies_temporal_question_with_when(self, handler):
        """Test detection of temporal questions with 'when'."""
        assert (
            handler._identify_question_type("When was the company founded?")
            == "temporal"
        )

    def test_identifies_temporal_question_with_year(self, handler):
        """Test detection of temporal questions with 'year'."""
        assert (
            handler._identify_question_type("In what year did the war begin?")
            == "temporal"
        )

    def test_identifies_temporal_question_with_date(self, handler):
        """Test detection of temporal questions with 'date'."""
        assert (
            handler._identify_question_type("What is the date of the event?")
            == "temporal"
        )

    def test_identifies_number_question_with_how_many(self, handler):
        """Test detection of number questions with 'how many'."""
        assert (
            handler._identify_question_type("How many employees work there?")
            == "number"
        )

    def test_identifies_number_question_with_how_much(self, handler):
        """Test detection of number questions with 'how much'."""
        assert (
            handler._identify_question_type("How much did it cost?") == "number"
        )

    def test_identifies_score_question(self, handler):
        """Test detection of score questions."""
        assert (
            handler._identify_question_type("What was the final score?")
            == "score"
        )

    def test_identifies_score_question_with_result(self, handler):
        """Test detection of score questions with 'result'."""
        assert (
            handler._identify_question_type("What was the result of the match?")
            == "score"
        )

    def test_identifies_dimension_question_with_height(self, handler):
        """Test detection of dimension questions with 'height'."""
        assert (
            handler._identify_question_type(
                "What is the height of the building?"
            )
            == "dimension"
        )

    def test_identifies_dimension_question_with_tall(self, handler):
        """Test detection of dimension questions with 'tall'."""
        assert (
            handler._identify_question_type("How tall is the tower?")
            == "dimension"
        )

    def test_identifies_dimension_question_with_length(self, handler):
        """Test detection of dimension questions with 'length'."""
        assert (
            handler._identify_question_type("What is the length of the bridge?")
            == "dimension"
        )

    def test_identifies_single_choice_question(self, handler):
        """Test detection of single choice questions."""
        assert (
            handler._identify_question_type("Which one of these is correct?")
            == "single_choice"
        )

    def test_returns_general_for_ambiguous_question(self, handler):
        """Test that ambiguous questions return 'general'."""
        assert (
            handler._identify_question_type("What happened in the meeting?")
            == "general"
        )

    def test_case_insensitive_detection(self, handler):
        """Test case-insensitive question type detection."""
        assert (
            handler._identify_question_type("WHAT IS THE FULL NAME?")
            == "full_name"
        )
        assert (
            handler._identify_question_type("WHERE DID IT HAPPEN?")
            == "location"
        )


class TestPrecisionExtractionRegexPatterns:
    """Tests for regex pattern matching."""

    @pytest.fixture
    def handler(self, mock_llm):
        """Create a PrecisionExtractionHandler."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        return PrecisionExtractionHandler(llm=mock_llm)

    def test_full_name_pattern_matches_two_words(self, handler):
        """Test full name pattern matches two-word names."""
        matches = handler.answer_patterns["full_name"].findall(
            "John Smith went to the store."
        )
        assert "John Smith" in matches

    def test_full_name_pattern_matches_three_words(self, handler):
        """Test full name pattern matches three-word names."""
        matches = handler.answer_patterns["full_name"].findall(
            "John Michael Smith was born."
        )
        assert "John Michael Smith" in matches

    def test_year_pattern_matches_1900s(self, handler):
        """Test year pattern matches 1900s years."""
        matches = handler.answer_patterns["year"].findall("Founded in 1995.")
        assert "1995" in matches

    def test_year_pattern_matches_2000s(self, handler):
        """Test year pattern matches 2000s years."""
        matches = handler.answer_patterns["year"].findall("Launched in 2024.")
        assert "2024" in matches

    def test_number_pattern_matches_integer(self, handler):
        """Test number pattern matches integers."""
        matches = handler.answer_patterns["number"].findall(
            "There are 42 employees."
        )
        assert "42" in matches

    def test_number_pattern_matches_decimal(self, handler):
        """Test number pattern matches decimals."""
        matches = handler.answer_patterns["number"].findall(
            "The cost is 19.99 dollars."
        )
        assert "19.99" in matches

    def test_dimension_pattern_matches_meters(self, handler):
        """Test dimension pattern matches meters."""
        matches = handler.answer_patterns["dimension"].findall(
            "The tower is 324 meters tall."
        )
        assert ("324", "meters") in matches

    def test_dimension_pattern_matches_feet(self, handler):
        """Test dimension pattern matches feet."""
        matches = handler.answer_patterns["dimension"].findall(
            "It is 100 feet high."
        )
        assert ("100", "feet") in matches

    def test_dimension_pattern_matches_kg(self, handler):
        """Test dimension pattern matches kilograms."""
        matches = handler.answer_patterns["dimension"].findall("Weighs 50 kg.")
        assert ("50", "kg") in matches

    def test_score_pattern_matches_hyphen_format(self, handler):
        """Test score pattern matches X-Y format."""
        matches = handler.answer_patterns["score"].findall(
            "Final score was 3-2."
        )
        assert ("3", "2") in matches

    def test_score_pattern_matches_dash_format(self, handler):
        """Test score pattern matches X–Y format (en-dash)."""
        matches = handler.answer_patterns["score"].findall(
            "Final score was 3–2."
        )
        assert ("3", "2") in matches


class TestPrecisionExtractionMethods:
    """Tests for individual extraction methods."""

    @pytest.fixture
    def handler(self, mock_llm):
        """Create a PrecisionExtractionHandler."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        return PrecisionExtractionHandler(llm=mock_llm)

    def test_extract_full_name_with_llm(self, handler, mock_llm):
        """Test _extract_full_name calls LLM."""
        mock_llm.invoke.return_value = Mock(
            content="Full name: John Michael Smith"
        )

        result = handler._extract_full_name(
            "John Smith was a scientist.", "Who is it?", "Sources"
        )

        mock_llm.invoke.assert_called_once()
        assert "John Michael Smith" in result

    def test_extract_full_name_fallback_to_longest(self, handler, mock_llm):
        """Test _extract_full_name falls back to longest name when LLM doesn't identify."""
        mock_llm.invoke.return_value = Mock(content="No full name identified")

        content = "John Michael Smith and John Smith were scientists."
        result = handler._extract_full_name(content, "Who is it?", "Sources")

        # Should pick the longer name
        assert "John Michael Smith" in result

    def test_extract_full_name_handles_exception(self, handler, mock_llm):
        """Test _extract_full_name handles exceptions gracefully."""
        mock_llm.invoke.side_effect = Exception("LLM error")

        result = handler._extract_full_name(
            "John Smith content", "Who?", "Sources"
        )

        # Should return original content
        assert "John Smith content" == result

    def test_extract_best_name_returns_most_frequent(self, handler):
        """Test _extract_best_name returns most frequent name."""
        content = "John Smith said hello. Smith was happy. John Smith left."

        result = handler._extract_best_name(content, "Who is it?", "Sources")

        assert "John Smith" in result

    def test_extract_dimension_with_llm(self, handler, mock_llm):
        """Test _extract_dimension calls LLM."""
        mock_llm.invoke.return_value = Mock(content="324 meters")

        result = handler._extract_dimension(
            "The tower is 324 meters tall.", "How tall?", "Sources"
        )

        mock_llm.invoke.assert_called_once()
        assert "324" in result

    def test_extract_dimension_fallback_to_pattern(self, handler, mock_llm):
        """Test _extract_dimension falls back to regex pattern."""
        mock_llm.invoke.return_value = Mock(content="No clear answer")

        content = "The tower is 324 meters tall."
        result = handler._extract_dimension(content, "How tall?", "Sources")

        # Should still extract the measurement
        assert "324" in result or "meters" in result or content == result

    def test_extract_score_with_multiple_scores(self, handler, mock_llm):
        """Test _extract_score when multiple scores present."""
        mock_llm.invoke.return_value = Mock(content="3-2")

        content = "Halftime was 1-1. Final score was 3-2."
        result = handler._extract_score(
            content, "What was the final score?", "Sources"
        )

        mock_llm.invoke.assert_called_once()
        assert "3-2" in result

    def test_extract_score_no_scores_returns_content(self, handler, mock_llm):
        """Test _extract_score returns content when no scores found."""
        content = "The game was exciting."
        result = handler._extract_score(
            content, "What was the score?", "Sources"
        )

        assert result == content

    def test_extract_temporal_with_llm(self, handler, mock_llm):
        """Test _extract_temporal calls LLM."""
        mock_llm.invoke.return_value = Mock(content="1998")

        content = "Founded in 1998, acquired in 2015."
        result = handler._extract_temporal(
            content, "When was it founded?", "Sources"
        )

        mock_llm.invoke.assert_called_once()
        assert "1998" in result

    def test_extract_temporal_no_years_returns_content(self, handler, mock_llm):
        """Test _extract_temporal returns content when no years found."""
        content = "It happened long ago."
        result = handler._extract_temporal(content, "When?", "Sources")

        assert result == content

    def test_extract_number_with_llm(self, handler, mock_llm):
        """Test _extract_number calls LLM."""
        mock_llm.invoke.return_value = Mock(content="42")

        content = "There are 42 employees and 10 managers."
        result = handler._extract_number(
            content, "How many employees?", "Sources"
        )

        mock_llm.invoke.assert_called_once()
        assert "42" in result

    def test_extract_single_answer_removes_alternatives(
        self, handler, mock_llm
    ):
        """Test _extract_single_answer removes comma/and/or alternatives."""
        mock_llm.invoke.return_value = Mock(
            content="Option A, Option B, and Option C"
        )

        result = handler._extract_single_answer(
            "Multiple options content", "Which one?", "Sources"
        )

        # Should only have first option
        assert "Option A" in result
        assert "Option B" not in result.split(".")[0]

    def test_extract_key_facts_from_previous_knowledge(self, handler, mock_llm):
        """Test _extract_key_facts extracts from previous knowledge."""
        mock_llm.invoke.return_value = Mock(
            content="Key facts: Founded 1998, Location: NYC"
        )

        result = handler._extract_key_facts(
            "Previous knowledge about the company", "name"
        )

        mock_llm.invoke.assert_called_once()
        assert len(result) <= 500

    def test_extract_key_facts_handles_exception(self, handler, mock_llm):
        """Test _extract_key_facts handles exception gracefully."""
        mock_llm.invoke.side_effect = Exception("LLM error")

        result = handler._extract_key_facts("Previous knowledge", "name")

        # Should return truncated previous knowledge
        assert "Previous knowledge" in result


class TestPrecisionExtractionAnalyzeInitial:
    """Tests for analyze_initial method."""

    def test_analyze_initial_returns_content_and_documents(
        self, mock_llm, sample_search_results
    ):
        """Test analyze_initial returns proper structure."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        handler = PrecisionExtractionHandler(llm=mock_llm)

        result = handler.analyze_initial(
            "What is the topic?", sample_search_results
        )

        assert "content" in result
        assert "documents" in result
        assert len(result["documents"]) == 3

    def test_analyze_initial_prompt_includes_question_type(
        self, mock_llm, sample_search_results
    ):
        """Test analyze_initial prompt includes detected question type."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        mock_llm.invoke.return_value = Mock(
            content="The full name is John Smith. [1]"
        )

        handler = PrecisionExtractionHandler(llm=mock_llm)

        handler.analyze_initial(
            "What is the full name of the founder?", sample_search_results
        )

        call_args = mock_llm.invoke.call_args_list[0][0][0]
        assert "full_name" in call_args

    def test_analyze_initial_prompt_includes_precision_instructions(
        self, mock_llm, sample_search_results
    ):
        """Test analyze_initial prompt includes precision extraction instructions."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        mock_llm.invoke.return_value = Mock(content="Answer [1]")

        handler = PrecisionExtractionHandler(llm=mock_llm)

        handler.analyze_initial("What is it?", sample_search_results)

        call_args = mock_llm.invoke.call_args_list[0][0][0]
        assert "PRECISION" in call_args
        assert "EXACT answer" in call_args

    def test_analyze_initial_applies_extraction_for_full_name(
        self, mock_llm, name_search_results
    ):
        """Test analyze_initial applies extraction for full_name questions."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        # First call returns LLM response, second call extracts name
        mock_llm.invoke.side_effect = [
            Mock(content="The person is John Smith."),
            Mock(content="Full name: John Michael William Smith"),
        ]

        handler = PrecisionExtractionHandler(llm=mock_llm)

        handler.analyze_initial("What is the full name?", name_search_results)

        # Should have called extraction
        assert mock_llm.invoke.call_count >= 1

    def test_analyze_initial_handles_string_response(
        self, mock_llm_string_response, sample_search_results
    ):
        """Test analyze_initial handles string LLM response."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        handler = PrecisionExtractionHandler(llm=mock_llm_string_response)

        result = handler.analyze_initial("What is it?", sample_search_results)

        assert "content" in result


class TestPrecisionExtractionAnalyzeFollowup:
    """Tests for analyze_followup method."""

    def test_analyze_followup_returns_content_and_documents(
        self, mock_llm, sample_search_results, sample_previous_knowledge
    ):
        """Test analyze_followup returns proper structure."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        handler = PrecisionExtractionHandler(llm=mock_llm)

        result = handler.analyze_followup(
            "Follow-up question?",
            sample_search_results,
            sample_previous_knowledge,
            nr_of_links=3,
        )

        assert "content" in result
        assert "documents" in result

    def test_analyze_followup_extracts_key_facts(
        self, mock_llm, sample_search_results, sample_previous_knowledge
    ):
        """Test analyze_followup extracts key facts from previous knowledge."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        mock_llm.invoke.return_value = Mock(content="Key facts extracted")

        handler = PrecisionExtractionHandler(llm=mock_llm)

        handler.analyze_followup(
            "What year was it founded?",
            sample_search_results,
            sample_previous_knowledge,
            nr_of_links=0,
        )

        # Should have called LLM for key facts extraction
        first_call = mock_llm.invoke.call_args_list[0][0][0]
        assert (
            "key facts" in first_call.lower() or "extract" in first_call.lower()
        )

    def test_analyze_followup_applies_nr_of_links_offset(
        self, mock_llm, sample_search_results, sample_previous_knowledge
    ):
        """Test analyze_followup applies nr_of_links offset to document indices."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        handler = PrecisionExtractionHandler(llm=mock_llm)

        result = handler.analyze_followup(
            "Question?",
            sample_search_results,
            sample_previous_knowledge,
            nr_of_links=10,
        )

        # First document should have index 11 (10 + 1)
        assert result["documents"][0].metadata["index"] == 11


class TestPrecisionExtractionEdgeCases:
    """Tests for edge cases and error handling."""

    def test_empty_search_results(self, mock_llm, empty_search_results):
        """Test handling of empty search results."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        handler = PrecisionExtractionHandler(llm=mock_llm)

        result = handler.analyze_initial("What is it?", empty_search_results)

        assert result["documents"] == []

    def test_string_search_results(self, mock_llm, string_search_results):
        """Test handling of string search results."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        handler = PrecisionExtractionHandler(llm=mock_llm)

        result = handler.analyze_initial("What is it?", string_search_results)

        assert result["documents"] == []

    def test_with_output_instructions(
        self, mock_llm, sample_search_results, settings_with_output_instructions
    ):
        """Test that output instructions are included in prompts."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        mock_llm.invoke.return_value = Mock(content="Response")

        handler = PrecisionExtractionHandler(
            llm=mock_llm, settings_snapshot=settings_with_output_instructions
        )

        handler.analyze_initial("What is it?", sample_search_results)

        call_args = mock_llm.invoke.call_args_list[0][0][0]
        assert "formal academic English" in call_args

    def test_general_question_type_no_extraction(
        self, mock_llm, sample_search_results
    ):
        """Test that general question type doesn't apply special extraction."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        mock_llm.invoke.return_value = Mock(content="General response")

        handler = PrecisionExtractionHandler(llm=mock_llm)

        result = handler.analyze_initial(
            "What happened?", sample_search_results
        )

        # Should just return the LLM response without extraction modifications
        assert result["content"] == "General response"

    def test_apply_precision_extraction_returns_content_for_unknown_type(
        self, mock_llm
    ):
        """Test _apply_precision_extraction returns content for unknown question type."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        handler = PrecisionExtractionHandler(llm=mock_llm)

        result = handler._apply_precision_extraction(
            "Original content", "Query?", "unknown_type", "Sources"
        )

        assert result == "Original content"


class TestPrecisionExtractionDimensionContextAwareness:
    """Tests for dimension extraction context awareness."""

    @pytest.fixture
    def handler(self, mock_llm):
        """Create a PrecisionExtractionHandler."""
        from local_deep_research.citation_handlers.precision_extraction_handler import (
            PrecisionExtractionHandler,
        )

        return PrecisionExtractionHandler(llm=mock_llm)

    def test_dimension_types_include_height(self, handler):
        """Test dimension type detection includes height keywords."""
        question_type = handler._identify_question_type(
            "What is the height of the building?"
        )
        assert question_type == "dimension"

    def test_dimension_types_include_size(self, handler):
        """Test dimension type detection includes size keywords."""
        question_type = handler._identify_question_type(
            "What is the size of the building?"
        )
        assert question_type == "dimension"

    def test_dimension_types_include_length(self, handler):
        """Test dimension type detection includes length keywords."""
        question_type = handler._identify_question_type(
            "What is the length of the bridge?"
        )
        assert question_type == "dimension"

    def test_dimension_types_include_width(self, handler):
        """Test dimension type detection includes width keywords."""
        question_type = handler._identify_question_type(
            "How wide is the river?"
        )
        assert question_type == "dimension"



import threading
import uuid
from functools import wraps
from typing import Any, Callable, Hashable, Tuple

from cachetools import cached, keys
from flask import current_app, g
from flask.ctx import AppContext
from loguru import logger

g_thread_local_store = threading.local()


def thread_specific_cache(*args: Any, **kwargs: Any) -> Callable:
    """
    A version of `cached()` that is local to a single thread. In other words,
    cache entries will only be valid in the thread where they were created.

    Args:
        *args: Will be forwarded to `cached()`.
        **kwargs: Will be forwarded to `cached()`.

    Returns:
        The wrapped function.

    """

    def _key_func(*args_: Any, **kwargs_: Any) -> Tuple[Hashable, ...]:
        base_hash = keys.hashkey(*args_, **kwargs_)

        if hasattr(g_thread_local_store, "thread_id"):
            # We already gave this thread a unique ID. Use that.
            thread_id = g_thread_local_store.thread_id
        else:
            # Give this thread a new unique ID.
            thread_id = uuid.uuid4().hex
            g_thread_local_store.thread_id = thread_id

        return (thread_id,) + base_hash

    return cached(*args, **kwargs, key=_key_func)


def thread_with_app_context(to_wrap: Callable) -> Callable:
    """
    Decorator that wraps the entry point to a thread and injects the current
    app context from Flask. This is useful when we want to use multiple
    threads to handle a single request.

    When using this wrapped function, `current_app.app_context()` should be
    passed as the first argument when initializing the thread.

    Args:
        to_wrap: The function to wrap.

    Returns:
        The wrapped function.

    """

    @wraps(to_wrap)
    def _run_with_context(
        app_context: AppContext | None, *args: Any, **kwargs: Any
    ) -> Any:
        if app_context is None:
            # Do nothing.
            return to_wrap(*args, **kwargs)

        with app_context:
            return to_wrap(*args, **kwargs)

    return _run_with_context


def thread_context() -> AppContext | None:
    """
    Pushes a new app context for a thread that is being spawned to handle the
    current request. Will copy all the global data from the current context.

    Returns:
        The new context, or None if no context is active.

    """
    # Copy global data.
    global_data = {}
    try:
        for key in g:
            global_data[key] = g.get(key)
    except TypeError:
        # Context is not initialized. Don't change anything.
        pass

    try:
        context = current_app.app_context()
    except RuntimeError:
        # Context is not initialized.
        logger.debug("No current app context, not passing to thread.")
        return None

    with context:
        for key, value in global_data.items():
            setattr(g, key, value)

    return context



"""
Tests for DownloadService.
"""

from unittest.mock import Mock, MagicMock


class TestDownloadServiceInit:
    """Tests for DownloadService initialization."""

    def test_init_creates_downloaders(self, mocker):
        """Initializes with list of downloaders."""
        # Mock the settings manager
        mock_settings = Mock()
        mock_settings.get_setting.return_value = "/tmp/test_library"
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_settings_manager",
            return_value=mock_settings,
        )

        # Mock Path.mkdir
        mocker.patch("pathlib.Path.mkdir")

        # Mock RetryManager
        mock_retry_manager = Mock()
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.RetryManager",
            return_value=mock_retry_manager,
        )

        from src.local_deep_research.research_library.services.download_service import (
            DownloadService,
        )

        service = DownloadService(username="test_user")

        assert service.username == "test_user"
        assert len(service.downloaders) > 0
        assert service.library_root == "/tmp/test_library"


class TestDownloadServiceUrlNormalization:
    """Tests for URL normalization."""

    def test_normalize_url_removes_protocol(self, mocker):
        """URL normalization removes http/https protocol."""
        # Create service with minimal initialization
        mock_settings = Mock()
        mock_settings.get_setting.return_value = "/tmp/test_library"
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_settings_manager",
            return_value=mock_settings,
        )
        mocker.patch("pathlib.Path.mkdir")
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.RetryManager"
        )

        from src.local_deep_research.research_library.services.download_service import (
            DownloadService,
        )

        service = DownloadService(username="test_user")

        url1 = service._normalize_url("https://example.com/path")
        url2 = service._normalize_url("http://example.com/path")

        assert url1 == url2
        assert not url1.startswith("http")

    def test_normalize_url_removes_www(self, mocker):
        """URL normalization removes www prefix."""
        mock_settings = Mock()
        mock_settings.get_setting.return_value = "/tmp/test_library"
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_settings_manager",
            return_value=mock_settings,
        )
        mocker.patch("pathlib.Path.mkdir")
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.RetryManager"
        )

        from src.local_deep_research.research_library.services.download_service import (
            DownloadService,
        )

        service = DownloadService(username="test_user")

        url1 = service._normalize_url("https://www.example.com/path")
        url2 = service._normalize_url("https://example.com/path")

        assert url1 == url2
        assert "www" not in url1

    def test_normalize_url_removes_trailing_slash(self, mocker):
        """URL normalization removes trailing slashes."""
        mock_settings = Mock()
        mock_settings.get_setting.return_value = "/tmp/test_library"
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_settings_manager",
            return_value=mock_settings,
        )
        mocker.patch("pathlib.Path.mkdir")
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.RetryManager"
        )

        from src.local_deep_research.research_library.services.download_service import (
            DownloadService,
        )

        service = DownloadService(username="test_user")

        url1 = service._normalize_url("https://example.com/path/")
        url2 = service._normalize_url("https://example.com/path")

        assert url1 == url2
        assert not url1.endswith("/")

    def test_normalize_url_sorts_query_params(self, mocker):
        """URL normalization sorts query parameters."""
        mock_settings = Mock()
        mock_settings.get_setting.return_value = "/tmp/test_library"
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_settings_manager",
            return_value=mock_settings,
        )
        mocker.patch("pathlib.Path.mkdir")
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.RetryManager"
        )

        from src.local_deep_research.research_library.services.download_service import (
            DownloadService,
        )

        service = DownloadService(username="test_user")

        url1 = service._normalize_url("https://example.com/path?b=2&a=1")
        url2 = service._normalize_url("https://example.com/path?a=1&b=2")

        assert url1 == url2


class TestDownloadServiceIsDownloadable:
    """Tests for _is_downloadable method."""

    def test_is_downloadable_pdf_extension(self, mocker):
        """Identifies .pdf URLs as downloadable."""
        mock_settings = Mock()
        mock_settings.get_setting.return_value = "/tmp/test_library"
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_settings_manager",
            return_value=mock_settings,
        )
        mocker.patch("pathlib.Path.mkdir")
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.RetryManager"
        )

        from src.local_deep_research.research_library.services.download_service import (
            DownloadService,
        )

        service = DownloadService(username="test_user")

        mock_resource = Mock()
        mock_resource.url = "https://example.com/paper.pdf"

        assert service._is_downloadable(mock_resource) is True

    def test_is_downloadable_arxiv_url(self, mocker):
        """Identifies arXiv URLs as downloadable."""
        mock_settings = Mock()
        mock_settings.get_setting.return_value = "/tmp/test_library"
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_settings_manager",
            return_value=mock_settings,
        )
        mocker.patch("pathlib.Path.mkdir")
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.RetryManager"
        )

        from src.local_deep_research.research_library.services.download_service import (
            DownloadService,
        )

        service = DownloadService(username="test_user")

        mock_resource = Mock()
        mock_resource.url = "https://arxiv.org/abs/2301.00001"

        assert service._is_downloadable(mock_resource) is True

    def test_is_downloadable_pubmed_pmc_url(self, mocker):
        """Identifies PubMed Central URLs as downloadable."""
        mock_settings = Mock()
        mock_settings.get_setting.return_value = "/tmp/test_library"
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_settings_manager",
            return_value=mock_settings,
        )
        mocker.patch("pathlib.Path.mkdir")
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.RetryManager"
        )

        from src.local_deep_research.research_library.services.download_service import (
            DownloadService,
        )

        service = DownloadService(username="test_user")

        mock_resource = Mock()
        mock_resource.url = "https://ncbi.nlm.nih.gov/pmc/articles/PMC1234567"

        assert service._is_downloadable(mock_resource) is True

    def test_is_downloadable_biorxiv_url(self, mocker):
        """Identifies bioRxiv URLs as downloadable."""
        mock_settings = Mock()
        mock_settings.get_setting.return_value = "/tmp/test_library"
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_settings_manager",
            return_value=mock_settings,
        )
        mocker.patch("pathlib.Path.mkdir")
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.RetryManager"
        )

        from src.local_deep_research.research_library.services.download_service import (
            DownloadService,
        )

        service = DownloadService(username="test_user")

        mock_resource = Mock()
        mock_resource.url = (
            "https://biorxiv.org/content/10.1101/2023.01.01.123456v1"
        )

        assert service._is_downloadable(mock_resource) is True

    def test_is_not_downloadable_regular_html(self, mocker):
        """Rejects regular HTML pages as not downloadable."""
        mock_settings = Mock()
        mock_settings.get_setting.return_value = "/tmp/test_library"
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_settings_manager",
            return_value=mock_settings,
        )
        mocker.patch("pathlib.Path.mkdir")
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.RetryManager"
        )

        from src.local_deep_research.research_library.services.download_service import (
            DownloadService,
        )

        service = DownloadService(username="test_user")

        mock_resource = Mock()
        mock_resource.url = "https://example.com/about.html"

        assert service._is_downloadable(mock_resource) is False


class TestDownloadServiceIsAlreadyDownloaded:
    """Tests for is_already_downloaded method."""

    def test_is_already_downloaded_true(self, mocker):
        """Returns True when URL is already downloaded."""
        mock_settings = Mock()
        mock_settings.get_setting.return_value = "/tmp/test_library"
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_settings_manager",
            return_value=mock_settings,
        )
        mocker.patch("pathlib.Path.mkdir")
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.RetryManager"
        )

        # Mock session and tracker
        mock_session = MagicMock()
        mock_session.__enter__ = Mock(return_value=mock_session)
        mock_session.__exit__ = Mock(return_value=False)

        mock_tracker = Mock()
        mock_tracker.is_downloaded = True
        mock_tracker.file_path = "pdfs/test.pdf"

        mock_session.query.return_value.filter_by.return_value.first.return_value = mock_tracker

        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_user_db_session",
            return_value=mock_session,
        )

        # Mock path exists
        mock_path = Mock()
        mock_path.exists.return_value = True
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_absolute_path_from_settings",
            return_value=mock_path,
        )

        from src.local_deep_research.research_library.services.download_service import (
            DownloadService,
        )

        service = DownloadService(username="test_user")

        is_downloaded, file_path = service.is_already_downloaded(
            "https://arxiv.org/pdf/2301.00001.pdf"
        )

        assert is_downloaded is True
        assert file_path is not None

    def test_is_already_downloaded_false_no_tracker(self, mocker):
        """Returns False when no tracker exists."""
        mock_settings = Mock()
        mock_settings.get_setting.return_value = "/tmp/test_library"
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_settings_manager",
            return_value=mock_settings,
        )
        mocker.patch("pathlib.Path.mkdir")
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.RetryManager"
        )

        # Mock session with no tracker
        mock_session = MagicMock()
        mock_session.__enter__ = Mock(return_value=mock_session)
        mock_session.__exit__ = Mock(return_value=False)
        mock_session.query.return_value.filter_by.return_value.first.return_value = None

        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_user_db_session",
            return_value=mock_session,
        )

        from src.local_deep_research.research_library.services.download_service import (
            DownloadService,
        )

        service = DownloadService(username="test_user")

        is_downloaded, file_path = service.is_already_downloaded(
            "https://arxiv.org/pdf/2301.00001.pdf"
        )

        assert is_downloaded is False
        assert file_path is None


class TestDownloadServiceGetDownloader:
    """Tests for _get_downloader method."""

    def test_get_downloader_for_arxiv(self, mocker):
        """Gets ArxivDownloader for arXiv URLs."""
        mock_settings = Mock()
        mock_settings.get_setting.return_value = "/tmp/test_library"
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_settings_manager",
            return_value=mock_settings,
        )
        mocker.patch("pathlib.Path.mkdir")
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.RetryManager"
        )

        from src.local_deep_research.research_library.services.download_service import (
            DownloadService,
        )
        from src.local_deep_research.research_library.downloaders import (
            ArxivDownloader,
        )

        service = DownloadService(username="test_user")

        downloader = service._get_downloader("https://arxiv.org/abs/2301.00001")

        assert downloader is not None
        assert isinstance(downloader, ArxivDownloader)

    def test_get_downloader_for_pubmed(self, mocker):
        """Gets PubMedDownloader for PubMed URLs."""
        mock_settings = Mock()
        mock_settings.get_setting.return_value = "/tmp/test_library"
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_settings_manager",
            return_value=mock_settings,
        )
        mocker.patch("pathlib.Path.mkdir")
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.RetryManager"
        )

        from src.local_deep_research.research_library.services.download_service import (
            DownloadService,
        )
        from src.local_deep_research.research_library.downloaders import (
            PubMedDownloader,
        )

        service = DownloadService(username="test_user")

        downloader = service._get_downloader(
            "https://pubmed.ncbi.nlm.nih.gov/12345678"
        )

        assert downloader is not None
        assert isinstance(downloader, PubMedDownloader)

    def test_get_downloader_for_pdf_url(self, mocker):
        """Gets DirectPDFDownloader for direct PDF URLs."""
        mock_settings = Mock()
        mock_settings.get_setting.return_value = "/tmp/test_library"
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_settings_manager",
            return_value=mock_settings,
        )
        mocker.patch("pathlib.Path.mkdir")
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.RetryManager"
        )

        from src.local_deep_research.research_library.services.download_service import (
            DownloadService,
        )
        from src.local_deep_research.research_library.downloaders import (
            DirectPDFDownloader,
        )

        service = DownloadService(username="test_user")

        downloader = service._get_downloader("https://example.com/paper.pdf")

        assert downloader is not None
        assert isinstance(downloader, DirectPDFDownloader)


class TestDownloadServiceTextExtraction:
    """Tests for text extraction methods."""

    def test_extract_text_from_pdf_success(self, mocker, mock_pdf_content):
        """Extracts text from valid PDF content."""
        mock_settings = Mock()
        mock_settings.get_setting.return_value = "/tmp/test_library"
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_settings_manager",
            return_value=mock_settings,
        )
        mocker.patch("pathlib.Path.mkdir")
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.RetryManager"
        )

        # Mock pdfplumber
        mock_pdf = MagicMock()
        mock_page = MagicMock()
        mock_page.extract_text.return_value = "Extracted text from page 1"
        mock_pdf.pages = [mock_page]
        mock_pdf.__enter__ = Mock(return_value=mock_pdf)
        mock_pdf.__exit__ = Mock(return_value=False)

        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.pdfplumber.open",
            return_value=mock_pdf,
        )

        from src.local_deep_research.research_library.services.download_service import (
            DownloadService,
        )

        service = DownloadService(username="test_user")

        text = service._extract_text_from_pdf(mock_pdf_content)

        assert text is not None
        assert "Extracted text from page 1" in text

    def test_extract_text_from_pdf_empty(self, mocker):
        """Returns None when PDF has no extractable text."""
        mock_settings = Mock()
        mock_settings.get_setting.return_value = "/tmp/test_library"
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_settings_manager",
            return_value=mock_settings,
        )
        mocker.patch("pathlib.Path.mkdir")
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.RetryManager"
        )

        # Mock pdfplumber with no text
        mock_pdf = MagicMock()
        mock_page = MagicMock()
        mock_page.extract_text.return_value = ""
        mock_pdf.pages = [mock_page]
        mock_pdf.__enter__ = Mock(return_value=mock_pdf)
        mock_pdf.__exit__ = Mock(return_value=False)

        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.pdfplumber.open",
            return_value=mock_pdf,
        )

        # Also mock PyPDF as fallback with no text
        mock_reader = MagicMock()
        mock_pypdf_page = MagicMock()
        mock_pypdf_page.extract_text.return_value = ""
        mock_reader.pages = [mock_pypdf_page]
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.PdfReader",
            return_value=mock_reader,
        )

        from src.local_deep_research.research_library.services.download_service import (
            DownloadService,
        )

        service = DownloadService(username="test_user")

        text = service._extract_text_from_pdf(b"%PDF-1.4 empty")

        assert text is None


class TestDownloadServiceQueueResearchDownloads:
    """Tests for queue_research_downloads method."""

    def test_queue_research_downloads_success(self, mocker):
        """Queues downloads for downloadable resources."""
        mock_settings = Mock()
        mock_settings.get_setting.return_value = "/tmp/test_library"
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_settings_manager",
            return_value=mock_settings,
        )
        mocker.patch("pathlib.Path.mkdir")
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.RetryManager"
        )

        # Mock session
        mock_session = MagicMock()
        mock_session.__enter__ = Mock(return_value=mock_session)
        mock_session.__exit__ = Mock(return_value=False)

        # Create mock resources
        mock_resource1 = Mock()
        mock_resource1.id = 1
        mock_resource1.url = "https://arxiv.org/abs/2301.00001"

        mock_resource2 = Mock()
        mock_resource2.id = 2
        mock_resource2.url = "https://example.com/page.html"

        mock_session.query.return_value.filter_by.return_value.all.return_value = [
            mock_resource1,
            mock_resource2,
        ]
        mock_session.query.return_value.filter_by.return_value.first.return_value = None

        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_user_db_session",
            return_value=mock_session,
        )

        # Mock get_default_library_id
        mocker.patch(
            "src.local_deep_research.database.library_init.get_default_library_id",
            return_value="default-lib-id",
        )

        from src.local_deep_research.research_library.services.download_service import (
            DownloadService,
        )

        service = DownloadService(username="test_user")

        queued_count = service.queue_research_downloads("research-123")

        # Only arxiv URL should be queued (HTML page is not downloadable)
        assert queued_count >= 0


class TestDownloadServiceDownloadResource:
    """Tests for download_resource method."""

    def test_download_resource_not_found(self, mocker):
        """Returns error when resource not found."""
        mock_settings = Mock()
        mock_settings.get_setting.return_value = "/tmp/test_library"
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_settings_manager",
            return_value=mock_settings,
        )
        mocker.patch("pathlib.Path.mkdir")
        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.RetryManager"
        )

        # Mock session with no resource
        mock_session = MagicMock()
        mock_session.__enter__ = Mock(return_value=mock_session)
        mock_session.__exit__ = Mock(return_value=False)
        mock_session.query.return_value.get.return_value = None

        mocker.patch(
            "src.local_deep_research.research_library.services.download_service.get_user_db_session",
            return_value=mock_session,
        )

        from src.local_deep_research.research_library.services.download_service import (
            DownloadService,
        )

        service = DownloadService(username="test_user")

        success, reason = service.download_resource(999)

        assert success is False
        assert "not found" in reason.lower()
